"""Fire modes and group class definitions."""

import enum
from typing import Dict, Final

from ..base import Cached
from ..census import Query
from ..models import FireGroupData, FireModeData
from ..proxy import InstanceProxy, SequenceProxy
from ..request import extract_payload

from .projectile import Projectile
from .states import PlayerState, PlayerStateGroup


class FireModeType(enum.IntEnum):
    """A type of fire mode.

    This is mostly used to group similar fire modes together when
    tabulating multiple weapons.
    """

    PROJECTILE = 0
    IRON_SIGHT = 1
    MELEE = 3
    TRIGGER_ITEM_ABILITY = 8
    THROWN = 12


class FireMode(Cached, cache_size=10, cache_ttu=3600.0):
    """A fire mode of a weapon.

    This class defines the bulk of a weapon's stats, such as reload
    times or accuracy.

    Note that this is not synonymous with in-game fire modes, these are
    handled by the :class:`FireGroup` class instead. Fire groups are
    also used to implement the auxiliary under-barrel fire modes.

    Implementation detail: This class wraps the ``fire_mode_2``
    collection, the regular ``fire_mode`` collection does not have a
    representation in the object model.
    """

    collection = 'fire_mode_2'
    data: FireModeData
    dataclass = FireModeData
    id_field = 'fire_mode_id'

    @property
    def type(self) -> FireModeType:
        """Return the type of fire mode as an enum."""
        return FireModeType(self.data.fire_mode_type_id)

    async def state_groups(self) -> Dict[PlayerState, PlayerStateGroup]:
        """Return the state-specific data for a fire mode."""
        collection: Final[str] = 'player_state_group_2'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field='player_state_group_id',
                       value=self.data.player_state_group_id)
        query.limit(10)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        states: Dict[PlayerState, PlayerStateGroup] = {}
        for group_data in data:
            group = PlayerStateGroup(**group_data)
            state = PlayerState(group.player_state_id)
            states[state] = group
        return states

    def projectile(self) -> InstanceProxy[Projectile]:
        """Return the projectile associated with this fire mode."""
        collection: Final[str] = 'fire_mode_to_projectile'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        join = query.create_join(Projectile.collection)
        join.set_fields(Projectile.id_field)
        return InstanceProxy(Projectile, query, client=self._client)


class FireGroup(Cached, cache_size=10, cache_ttu=60.0):
    """Links multiple fire modes into a group.

    Fire groups are comparable to the in-game fire modes, such as
    burst, semi auto or fully automatic. They are also used to
    implement auxiliary fire modes such as under-barrel launchers.
    """

    collection = 'fire_group'
    data: FireGroupData
    dataclass = FireGroupData
    id_field = 'fire_group_id'

    def fire_modes(self) -> SequenceProxy[FireMode]:
        """Return the fire modes in the fire group."""
        collection: Final[str] = 'fire_group_to_fire_mode'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(20)
        join = query.create_join(FireMode.collection)
        join.set_fields(FireMode.id_field)
        return SequenceProxy(FireMode, query, client=self._client)



"""Data classes for :mod:`auraxium.ps2.character`."""

from typing import Optional

from ..base import Ps2Data
from ..types import LocaleData

__all__ = [
    'CharacterAchievement',
    'CharacterData',
    'CharacterDirective',
    'TitleData'
]


class CharacterAchievement(Ps2Data):
    """Data container for a character's achievement status.

    Attributes:
        character_id: The ID of the character for this entry.
        achievement_id: The ID of the achievement for this entry.
        earned_count: How often the character has earned the given
            achievement.
        start: The UTC timestamp the character started progression
            towards this achievement at.
        start_date: Human-readable version of :attr:`start`.
        finish: The time the character completed this achievement. Only
            valid for one-time achievements such as medals.
        finish_date: Human-readable version of :attr:`finish`.
        last_save: The last time the character gained this
            achievement at as a UTC timestamp.
        last_save_date: Human-readable version of :attr:`last_save`.

        """

    character_id: int
    achievement_id: int
    earned_count: int
    start: int
    start_date: str
    finish: int
    finish_date: str
    last_save: int
    last_save_date: str


class CharacterData(Ps2Data):
    """Data class for :class:`auraxium.ps2.character.Character`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        character_id: The unique name of the character.
        name: The name of the player.
        faction_id: The faction the character belongs to.
        head_id: The head model for this character.
        title_id: The current title selected for this character. May be
            zero.
        times: Play and login time data for the character.
        certs: Certification data for the character.
        battle_rank: Battle rank data for the character.
        profile_id: The last profile the character used.
        daily_ribbon: Daily ribbon data for the character.
        prestige_level: The ASP rank of the character.

    """

    class BattleRank(Ps2Data):
        """Object representation of the "battle_rank" sub-key.

        Attributes:
            value: The current battle rank.
            percent_to_next: The progress to the next battle rank.

        """

        value: int
        percent_to_next: float

    class Certs(Ps2Data):
        """Object representation of the "certs" sub-key.

        Attributes:
            earned_points: Certification points the player has ever
                earned.
            gifted_points: Certification points the player was gifted
                through events.
            spent_points: Certification points the character has spent.
            available_points: The current certification point balance
                of the character.
            percent_to_next: The progress to the next certification
                point for the character.

            """

        earned_points: int
        gifted_points: int
        spent_points: int
        available_points: int
        percent_to_next: float

    class DailyRibbon(Ps2Data):
        """Object representation of the "daily_ribbon" sub-key.

        Attributes:
            count: The number of daily ribbon bonuses available.
            time: (Not yet documented)
            date: Human-readable version of :attr:`time`.

        """

        count: int  # type: ignore
        time: Optional[int] = None
        date: Optional[str] = None

    class Name(Ps2Data):
        """Object representation of the "name" sub-key.

        Attributes:
            first: The name of the character.
            first_lower: Lowercase version of :attr:`first`.

        """

        first: str
        first_lower: str

    class Times(Ps2Data):
        """Object representation of the "times" sub-key.

        Attributes:
            creation: The time the character was created.
            creation_date: Human-readable version of :attr:`creation`.
            last_save: The last time the character was updated. This
                roughly matches the last time the character logged out.
            last_save_date: Human-readable version of
                :attr:`last_save`.
            last_login: The last time the character logged in.
            last_login_date: Human-readable version of
                :attr:`last_login_date`.
            login_count: The number of times the character has logged
                in.
            minutes_played: The total number of minutes this character
                was logged into PS2.

            """

        creation: int
        creation_date: str
        last_save: int
        last_save_date: str
        last_login: int
        last_login_date: str
        login_count: int
        minutes_played: int

    character_id: int
    name: Name
    faction_id: int
    head_id: int
    title_id: int
    times: Times
    certs: Certs
    battle_rank: BattleRank
    profile_id: int
    daily_ribbon: DailyRibbon
    prestige_level: int


class CharacterDirective(Ps2Data):
    """Data container for a character's directive status.

    Attributes:
        character_id: The ID of the character for this entry.
        directive_tree_id: The ID of the directive tree for this entry.
        directive_id: The ID of the directive for this entry.
        completion_time: The time the character completed this
            directive.
        completion_time_date: Human-readable version of
            :attr:`completion_time`.

    """

    character_id: int
    directive_tree_id: int
    directive_id: int
    completion_time: int
    completion_time_date: str


class TitleData(Ps2Data):
    """Data class for :class:`auraxium.ps2.character.Title`.

    .. important::
        Unlike most other forms of API data, the ID used by titles is
        **not** unique.

        This is due to the ASP system re-using the same title IDs while
        introducing a different name ("A.S.P. Operative" for ``en``).

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        title_id: The ID of this title.
        name: The localised name of this title.

    """

    title_id: int
    name: LocaleData



"""Projectile and flight type class definitions."""

import enum

from ..base import Cached
from ..models import ProjectileData


class ProjectileFlightType(enum.IntEnum):
    """A projectile flight type."""

    BALLISTIC = 1
    TRUE_BALLISTIC = 3
    DYNAMIC = 9
    PROXIMITY_DETONATE = 10


class Projectile(Cached, cache_size=100, cache_ttu=60.0):
    """A projectile fired by a fire mode."""

    collection = 'projectile'
    data: ProjectileData
    dataclass = ProjectileData
    id_field = 'projectile_id'

    @property
    def flight_type(self) -> ProjectileFlightType:
        """Return the flight type of the projectile."""
        return ProjectileFlightType(self.data.projectile_flight_type_id)



"""Currency class definition."""

from ..base import Cached
from ..models import CurrencyData


class Currency(Cached, cache_size=10, cache_ttu=3600.0):
    """A currency obtainable by characters."""

    collection = 'currency'
    data: CurrencyData
    dataclass = CurrencyData
    id_field = 'currency_id'



"""Data classes for :mod:`auraxium.ps2.zone`."""

from ..base import Ps2Data
from ..types import LocaleData

__all__ = [
    'ZoneData'
]


class ZoneData(Ps2Data):
    """Data class for :class:`auraxium.ps2.zone.Zone`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        zone_id: The unique ID of this zone.
        code: The internal code used to represent this zone.
        hex_size: The map size for this zone.
        name: The localised name of this zone.
        description: The localised name of this zone.

    """

    zone_id: int
    code: str
    hex_size: int
    name: LocaleData
    description: LocaleData



"""Base classes for the Auraxium object model.

These classes define shared functionality required by all object
representations of API data, and defines the basic class hierarchy used
throughout the PlanetSide 2 object model.
"""

import abc
import dataclasses
import logging
from typing import (Any, ClassVar, Dict, List, Optional, Type, TYPE_CHECKING,
                    TypeVar, Union)

import pydantic

from .cache import TLRUCache
from .census import Query
from .errors import PayloadError, NotFoundError
from .request import extract_payload, extract_single
from .types import CensusData

if TYPE_CHECKING:
    # This is only imported during static type checking to resolve the 'Client'
    # forward reference. This avoids a circular import at runtime.
    from .client import Client

__all__ = [
    'Ps2Data',
    'Ps2Object',
    'Cached',
    'Named'
]

AnyT = TypeVar('AnyT')
CachedT = TypeVar('CachedT', bound='Cached')
NamedT = TypeVar('NamedT', bound='Named')
Ps2DataT = TypeVar('Ps2DataT', bound='Ps2Data')
Ps2ObjectT = TypeVar('Ps2ObjectT', bound='Ps2Object')
log = logging.getLogger('auraxium.ps2')


class Ps2Data(pydantic.BaseModel, metaclass=abc.ABCMeta):
    """Base class for PlanetSide 2 data classes.

    This class is based on :class:`pydantic.BaseModel` and will
    automatically cast any kwargs provided into the set of attributes
    defined for the current subclass.

    Extraneous kwargs are silently discarded. Any values equal to the
    string ``'NULL'`` are converted to ``None`` before the parsing is
    performed.

    This does support compound type declarations like
    :class:`typing.Union` or :class:`typing.Optional`.
    """

    class Config:
        """Pydantic model configuration.

        This inner class is used to namespace the pydantic
        configuration options.
        """
        allow_mutation = False
        anystr_strip_whitespace = True

    @pydantic.validator('*', pre=True)
    def convert_null(cls: Type['Ps2Data'], value: AnyT) -> Optional[AnyT]:
        """Handle NULL string return values.

        This converts any NULL strings to equal ``None`` instead.

        By default, the API will omit any NULL fields in the response,
        unless the ``c:includeNull`` flag is set.
        This value being a string would break type annotations like
        ``Optional[int] = None``, which is why they are silently
        converted into ``None`` before any parsing takes place.
        """
        # pylint: disable=no-self-use,no-self-argument
        if value == 'NULL':
            return None
        return value


class FallbackMixin:
    """A mixin class used to provide hard-coded fallback instances.

    Some collections are out of date and do not contain all required
    data. This mixin provides a hook to insert this missing data into
    data types while not causing issues if the API ends up being
    updated to include these missing types.
    """

    _fallback: ClassVar[Dict[int, CensusData]]


class Ps2Object(metaclass=abc.ABCMeta):
    """Common base class for all PS2 object representations.

    This requires that subclasses overwrite the
    :attr:`Ps2Object.collection` and :attr:`Ps2Object.id_field` names,
    which are used to tie the class to its corresponding API
    counterpart.

    """

    collection: ClassVar[str] = 'bogus'
    dataclass: ClassVar[Type[Ps2Data]]
    id_field: ClassVar[str] = 'bogus_id'

    def __init__(self, data: CensusData, client: 'Client') -> None:
        """Initialise the object.

        This sets the object's :attr:`~Ps2Object.id` attribute and
        populates the instance using the provided payload.

        Arguments:
            data: The census response dictionary to populate the
                object with.
            client (optional): The client object to use for requests
                performed via this object. Defaults to ``None``.

        """
        id_ = int(data[self.id_field])
        log.debug('Instantiating <%s:%d> using payload: %s',
                  self.__class__.__name__, id_, data)
        self.id = id_  # pylint: disable=invalid-name
        self._client = client
        try:
            self.data = self.dataclass(**data)
        except pydantic.ValidationError as err:
            raise PayloadError(
                f'Unable to instantiate {self.__class__.__name__} instance '
                f'from given payload: {err}', data) from err

    def __eq__(self, value: Any) -> bool:
        if not isinstance(value, self.__class__):
            return False
        return self.id == value.id

    def __hash__(self) -> int:
        return hash((self.__class__, self.id))

    def __repr__(self) -> str:
        """Return the unique string representation of this object.

        This will take the form of ``<Class:id>``, e.g.
        ``<Weapon:108>``.

        Returns:
            A string representing the object.

        """
        return f'<{self.__class__.__name__}:{self.id}>'

    @classmethod
    async def count(cls: Type[Ps2ObjectT], client: 'Client',
                    **kwargs: Any) -> int:
        """Return the number of items matching the given terms.

        Arguments:
            client: The client through which to perform the request.
            **kwargs: Any number of query filters to apply.

        Returns:
            The number of entries entries.

        """
        service_id = 's:example' if client is None else client.service_id
        query = Query(cls.collection, service_id=service_id, **kwargs)
        result = await client.request(query, verb='count')
        try:
            return int(result['count'])
        except KeyError as err:
            raise PayloadError(
                'Missing key "count" in API response', result) from err
        except ValueError as err:
            raise PayloadError(
                f'Invalid count: {result["count"]}', result) from err

    @classmethod
    async def find(cls: Type[Ps2ObjectT], results: int = 10, *,
                   offset: int = 0, promote_exact: bool = False,
                   check_case: bool = True, client: 'Client',
                   **kwargs: Any) -> List[Ps2ObjectT]:
        """Return a list of entries matching the given terms.

        This returns up to as many entries as indicated by the results
        argument. Note that it may be fewer if not enough matches are
        found.

        Arguments:
            results (optional): The maximum number of results. Defaults
                to ``10``.
            offset (optional): The number of entries to skip. Useful
                for paginated views. Defaults to ``0``.
            promote_exact (optional): If enabled, exact matches to
                non-exact searches will always come first in the return
                list. Defaults to ``False``.
            check_case (optional): Whether to check case when comparing
                strings. Note that case-insensitive searches are much
                more expensive. Defaults to ``True``.
            client: The client through which to perform the request.
            **kwargs: Any number of filters to apply.

        Returns:
            A list of matching entries.

        """
        service_id = 's:example' if client is None else client.service_id
        query = Query(cls.collection, service_id=service_id, **kwargs)
        query.limit(results)
        if offset > 0:
            query.offset(offset)
        query.exact_match_first(promote_exact).case(check_case)
        matches = await client.request(query)
        return [cls(i, client=client) for i in extract_payload(
            matches, cls.collection)]

    @classmethod
    async def get(cls: Type[Ps2ObjectT], client: 'Client',
                  check_case: bool = True, **kwargs: Any
                  ) -> Optional[Ps2ObjectT]:
        """Return the first entry matching the given terms.

        Like :meth:`Ps2Object.get()`, but will only return one item.

        Arguments:
            client: The client through which to perform the request.
            check_case (optional): Whether to check case when comparing
                strings. Note that case-insensitive searches are much
                more expensive. Defaults to ``True``.

        Returns:
            A matching entry, or None if not found.

        """
        results = await cls.find(client=client, results=1,
                                 check_case=check_case, **kwargs)
        if results:
            return results[0]
        return None

    @classmethod
    async def get_by_id(cls: Type[Ps2ObjectT], id_: int, *, client: 'Client'
                        ) -> Optional[Ps2ObjectT]:
        """Retrieve an object by its unique Census ID.

        Arguments:
            id_: The unique ID of the object.
            client: The client through which to perform the request.

        Returns:
            The entry with the matching ID, or None if not found.

        """
        filters: CensusData = {cls.id_field: id_}
        results = await cls.find(client=client, results=1, **filters)

        # Check for FallbackMixin compatibility
        if hasattr(cls, '_fallback'):
            # pylint: disable=no-member
            fallbacks: Dict[int, CensusData] = cls._fallback  # type: ignore
            log.debug('Fallback attribute found for type "%s", checking ID...',
                      cls.__name__)
            if (data := fallbacks.get(id_)) is not None:
                log.debug('Instantiating "%s" with ID %d through local copy',
                          cls.__name__, id_)

                if results:
                    # Log the fact that the local copy is not required
                    log.info('Type "%s" provides a local fallback for ID %d '
                             'despite this type being available on-line',
                             cls.__name__, id_)
                    return results[0]

                # Return a locally instantiated copy
                return cls(data, client=client)

            log.debug('No matching fallback instance found for ID %d', id_)

        elif results:
            # If no fallback value was provided, return the first item found
            # as normal
            return results[0]

        return None

    def query(self) -> Query:
        """Return a query from the current object.

        This is a utility method targeted at advanced users and
        developers. It is generally not required for most use cases.
        """
        return Query(collection=self.collection,
                     service_id=self._client.service_id)


class Cached(Ps2Object, metaclass=abc.ABCMeta):
    """Base class for cacheable data types.

    This generates a cache for each subclass that allows the storage
    and retrieval of objects by ID. This cache may be customised using
    keyword arguments as part of the class definition.

    This customisation is done via two parameters: the cache size and
    the TTU.

    The cache size defines the maximum number of items the cache may
    bold before it will discard the least recently used item for every
    new item added.

    The TTU (time-to-use) will independently discard items that are
    older than the given number of seconds to ensure data does not go
    too far out of date.

    """

    _cache: ClassVar[TLRUCache[int, Any]]

    def __init__(self, data: CensusData, client: 'Client') -> None:
        """Initialise the cached object.

        After initialising this object via the parent class's
        initialiser, this adds the current class to the cache.

        Arguments:
            data: The API response to instantiate the object from.
            client: The client used to retrieve the object.

        """
        super().__init__(data=data, client=client)
        self._cache.add(self.id, self)

    @classmethod
    def __init_subclass__(cls: Type[CachedT], cache_size: int,
                          cache_ttu: float = 0.0) -> None:
        """Initialise a cacheable subclass.

        This sets up the TLRU cache for the given subclass using the
        keyword arguments provided in the class definitions.

        Arguments:
            cache_size: The maximum number of items in the cache. Once
                the cache reaches this number of items, it will delete
                the  least recently used item for every new item added.
            cache_ttu (optional): The time-to-use for cache items. If
                an item is older than  TTU allows, it will be
                re-fetched regardless of how often it is accessed.
                Defaults to ``0.0``.

        """
        super().__init_subclass__()
        log.debug('Setting up cache for %s (size: %d, ttu: %.1f sec.)',
                  cls.__name__, cache_size, cache_ttu)
        cls._cache = TLRUCache(size=cache_size, ttu=cache_ttu,
                               name=f'{cls.__name__}_Cache')

    @classmethod
    def alter_cache(cls, size: int, ttu: Optional[float] = None) -> None:
        """Modify the class cache to use a new size and TTU.

        This will update and clear the cache for the current class.
        This allows customisation of the class depending on your
        use-case.

        Arguments:
            size: The new cache size.
            ttu (optional): The new item TTU. Defaults to ``None``.

        Raises:
            ValueError: Raised if the size is less than 1.

        """
        if size < 1:
            raise ValueError(f'{size} is not a valid cache size')
        cls._cache.clear()
        cls._cache.size = size
        if ttu is not None:
            cls._cache.ttu = ttu

    @classmethod
    def _check_cache(cls: Type[CachedT], id_: int) -> Optional[CachedT]:
        """Attempt to restore an item from the cache.

        If the item cannot be found, ``None`` will be returned instead.

        Arguments:
            id_: The unique identifier the item is cached by.

        Returns:
            An existing instance if found, or ``None`` if the object
            has not been retrieved before or expired.

        """
        return cls._cache.get(id_)

    @classmethod
    async def get_by_id(cls: Type[CachedT], id_: int, *,  # type: ignore
                        client: 'Client') -> Optional[CachedT]:
        """Retrieve an object by by ID.

        This query uses caches and might return an existing instance if
        the object has been recently retrieved.

        Arguments:
            id_: The unique id of the object.
            client: The client through which to perform the request.

        Returns:
            The object matching the given ID or ``None`` if no match
            was found.

        """
        log.debug('<%s:%d> requested', cls.__name__, id_)
        if (instance := cls._cache.get(id_)) is not None:
            log.debug('%r restored from cache', instance)
            return instance  # type: ignore
        log.debug('<%s:%d> not cached, generating API query...',
                  cls.__name__, id_)
        return await super().get_by_id(id_, client=client)  # type: ignore


class Named(Cached, cache_size=0, cache_ttu=0.0, metaclass=abc.ABCMeta):
    """Mix-in class for named objects.

    This extends the functionality provided by :class:`Cached` to also
    cache objects retrieved via :meth:`Named.get_by_name()`. The cache
    will also store the locale used for the request.

    """

    _cache: ClassVar[TLRUCache[Union[int, str], Any]]  # type: ignore

    def __init__(self, *args: Any, locale: Optional[str] = None,
                 **kwargs: Any) -> None:
        """Initialise the named object.

        This sets the object's id attribute and adds it to the cache.

        Arguments:
            locale: The locale under which to cache this object.
            *args: Any extra positional arguments are forwarded to the
                :class:`Cached` class's initialiser.
            **kwargs: Any keyword arguments are forwarded to the
                :class:`Cached` class's initialiser.

        """
        super().__init__(*args, **kwargs)
        if locale is not None:
            key = f'{locale}_{self.name(locale=locale).lower()}'
            self._cache.add(key, self)

    def __repr__(self) -> str:
        """Return the unique string representation of the faction.

        This will take the form of ``<class:id:name>``, e.g.
        ``<Item:2:NC4 Mag-Shot>``.

        Returns:
            A string representing the object.

        """
        return (f'<{self.__class__.__name__}:{self.id}:'
                f'\'{self.name(locale="en")}\'>')

    def __str__(self) -> str:
        """Return the string representation of this object.

        This calls the :meth:``Named.name()`` method for the English
        locale.

        Returns:
            A string representation of the object.

        """
        return self.name(locale='en')

    @classmethod
    async def get_by_name(cls: Type[NamedT], name: str, *, locale: str = 'en',
                          client: 'Client') -> Optional[NamedT]:
        """Retrieve an object by its unique name.

        If the same query has been performed recently, it may be
        restored from cache instead.

        This query is always case-insensitive.

        Arguments:
            name: The name to search for.
            locale (optional): The locale of the search key. Defaults
                to ``'en'``.
            client: The client through which to perform the request.

        Returns:
            The entry with the matching name, or ``None`` if not found.

        """
        key = f'{locale}_{name.lower()}'
        log.debug('%s "%s"[%s] requested', cls.__name__, name, locale)
        if (instance := cls._cache.get(key)) is not None:
            log.debug('%r restored from cache', instance)
            return instance  # type: ignore
        log.debug('%s "%s"[%s] not cached, generating API query...',
                  cls.__name__, name, locale)
        query = Query(cls.collection, service_id=client.service_id)
        query.case(False).add_term(field=f'name.{locale}', value=name)
        payload = await client.request(query)
        try:
            payload = extract_single(payload, cls.collection)
        except NotFoundError:
            return None
        return cls(payload, locale=locale, client=client)

    def name(self, locale: str = 'en') -> str:
        """Return the localised name of the object.

        Some subclasses may not have a localised name field. In these
        cases, the ``locale`` argument will be ignored.

        Arguments:
            locale (optional): The locale identifier to return.
                Defaults to ``'en'``.

        Raises:
            ValueError: Raised if the given locale is unknown.

        Returns:
            The localised name of the object.

        """
        data = self.data
        assert hasattr(data, 'name')
        try:
            return str(getattr(data.name, locale))  # type: ignore
        except AttributeError as err:
            raise ValueError(f'Invalid locale: {locale}') from err


class ImageMixin(Ps2Object, metaclass=abc.ABCMeta):
    """A mixin class for types supporting image access."""

    def image(self) -> str:
        """Return the default image for this type."""
        image_id: int = self.data.image_id  # type: ignore
        return self._image_url(image_id)

    @staticmethod
    def _image_url(image_id: int) -> str:
        """Return the URL for a given image ID."""
        url = 'https://census.daybreakgames.com/files/ps2/images/static/'
        return url + f'{image_id}.png'


@dataclasses.dataclass(frozen=True)
class ImageData:
    """Mixin dataclass for types supporting image access."""

    image_id: Optional[int] = None
    image_set_id: Optional[int] = None
    image_path: Optional[str] = None



"""Ability and ability type class definitions."""

from ..base import Cached
from ..census import Query
from ..models import ZoneEffectData, ZoneEffectTypeData
from ..proxy import InstanceProxy

from .ability import Ability


class ZoneEffectType(Cached, cache_size=20, cache_ttu=60.0):
    """A type of zone effect.

    This class mostly specifies the purpose of any generic parameters.
    """

    collection = 'zone_effect_type'
    data: ZoneEffectTypeData
    dataclass = ZoneEffectTypeData
    id_field = 'zone_effect_type_id'


class ZoneEffect(Cached, cache_size=10, cache_ttu=60.0):
    """An effect or buff applied by a zone.

    Access the corresponding
    :class:`auraxium.ps2.zone_effect.ZoneEffectType` instance via the
    :meth:`type` method for information on generic parameters.
    """

    collection = 'zone_effect'
    data: ZoneEffectData
    dataclass = ZoneEffectData
    id_field = 'zone_effect_id'

    def ability(self) -> InstanceProxy[Ability]:
        """Return the ability associated with this zone effect."""
        query = Query(Ability.collection, service_id=self._client.service_id)
        query.add_term(field=Ability.id_field, value=self.data.ability_id)
        return InstanceProxy(Ability, query, client=self._client)

    def type(self) -> InstanceProxy[ZoneEffectType]:
        """Return the type of this zone effect."""
        query = Query(
            ZoneEffectType.collection, service_id=self._client.service_id)
        query.add_term(
            field=ZoneEffectType.id_field, value=self.data.zone_effect_type_id)
        return InstanceProxy(ZoneEffectType, query, client=self._client)



"""Data classes for :mod:`auraxium.ps2.armour`."""

from typing import Optional

from ..base import Ps2Data

__all__ = [
    'ArmourInfoData'
]


class ArmourInfoData(Ps2Data):
    """Data class for :class:`auraxium.ps2.armour.ArmorInfo`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        armor_info_id: The unique ID of this entry.
        armor_facing_id: The enum value the facing direction this entry
            provides armour data for.
        armor_percent: Damage reduction in percent.
        armor_amount: A flat damage absorption applied to the damage
            effect; generally unused.
        description: A description for this entry.

    """

    armor_info_id: int
    armor_facing_id: int
    armor_percent: int
    armor_amount: Optional[int] = None
    description: str



"""Achievement class definition.

Achievements include weapon medals and service ribbons.
"""

from ..base import ImageMixin, Named
from ..models import AchievementData


class Achievement(Named, ImageMixin, cache_size=50, cache_ttu=60.0):
    """An achievement a player may pursue.

    Achievements include weapon medals and service ribbons.
    """

    collection = 'achievement'
    data: AchievementData
    dataclass = AchievementData
    id_field = 'achievement_id'



"""A Python wrapper for the PlanetSide 2 API.

This module provides an intuitive, object-oriented interface for using
the Census API. It targets interactive or medium traffic use-cases such
as Discord bots, player and outfit trackers, or aggregation of
historical data for AI purposes.

This is explicitly not targeting higher traffic use cases like website
back-ends, experiment at your own risk.

For additional information, bug reports or feature requests, visit the
project's repository at https://github.com/leonhard-s/auraxium.
"""

from . import census
from . import errors
from . import ps2
from .base import Cached, Named, Ps2Object
from .client import Client
from .event import Event, EventClient, EventType, Trigger

__all__ = [
    'Cached',
    'census',
    'Client',
    'errors',
    'Event',
    'EventClient',
    'EventType',
    'Named',
    'ps2',
    'Ps2Object',
    'Trigger'
]

__author__ = 'Leonhard S.'
__version__ = '0.1.0a2'



"""World class definition."""

import datetime
from typing import Any, Final, List, Optional, Tuple, Union

from ..base import Named
from ..census import Query
from ..client import Client
from ..models import WorldData
from ..request import extract_payload, extract_single
from ..types import CensusData

from .zone import Zone


class World(Named, cache_size=20, cache_ttu=3600.0):
    """A world (or server) in the game."""

    collection = 'world'
    data: WorldData
    dataclass = WorldData
    id_field = 'world_id'

    async def events(self, **kwargs: Any) -> List[CensusData]:
        """Return events for this world.

        This provides a REST endpoint for past alerts (MetagameEvent)
        and facility captures (FacilityCapture).

        This is always limited to at most 1000 return values. Use the
        begin and end parameters to poll more data.
        """
        collection: Final[str] = 'world_event'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(1000)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection=collection)
        return data

    @classmethod
    async def get_by_name(cls, name: str, *, locale: str = 'en',
                          client: Client) -> Optional['World']:
        """Retrieve a world by name.

        This query is always case-insensitive.
        """
        # NOTE: The world collection may only be queried by the world_id field
        # due to API limitations. This method works around this by first
        # retrieving all worlds, then looking the returned list up by name.
        worlds = await cls.find(20, client=client)
        name = name.lower()
        for world in worlds:
            if world.name(locale=locale).lower() == name:
                return world
        return None

    async def map(self, zone: Union[int, Zone],
                  *args: Union[int, Zone]) -> List[CensusData]:
        """Return the map status of a given zone."""
        collection: Final[str] = 'map'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        zone_ids: List[int] = [zone if isinstance(zone, int) else zone.id]
        zone_ids.extend(z if isinstance(z, int) else z.id for z in args)
        value = ','.join(str(z) for z in zone_ids)
        query.add_term(field='zone_ids', value=value)
        query.limit(3000)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection=collection)
        return data

    async def status(self) -> Tuple[str, datetime.datetime]:
        """Return the online status for the world.

        This returns a tuple consisting of the reported server state
        (e.g. "locked", "low", or "high", the latter referring to
        population numbers), and the last time this value was updated.
        """
        query = Query('game_server_status', namespace='global',
                      service_id=self._client.service_id, game_code='ps2',
                      name=f'^{self.data.name.en}')
        payload = await self._client.request(query)
        data = extract_single(payload, 'game_server_status')
        status = str(data['last_reported_state'])
        last_updated = int(data['last_reported_time'])
        return status, datetime.datetime.utcfromtimestamp(last_updated)



"""Data classes for :mod:`auraxium.ps2.weapon`."""

from typing import Optional

from ..base import Ps2Data
from ..types import LocaleData

__all__ = [
    'WeaponAmmoSlot',
    'WeaponData',
    'WeaponDatasheet'
]


class WeaponAmmoSlot(Ps2Data):
    """Data class for weapon ammo slot data.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        weapon_id: The ID of the associated :class:`Weapon`.
        weapon_slot_index: The position of the ammo type in the list of
            ammo types.
        clip_size: The clip size for this ammo type.
        capacity: The maximum amount of ammo that can be held with this
            ammo type.
        refill_ammo_rate: The amount of ammo being replenished every
            ammo refill tick.
        refill_ammo_delay_ms: The time between two ammo replenishment
            ticks.

    """

    weapon_id: int
    weapon_slot_index: int
    clip_size: int
    capacity: int
    refill_ammo_rate: Optional[int] = None
    refill_ammo_delay_ms: Optional[int] = None


class WeaponData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.Weapon`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        weapon_id: The unique ID of this weapon.
        weapon_group_id: Used to group upgradable weapons together and
            track them as a single entity, such as the Infiltrator's
            Recon Dart Device or the Engineer's Repair Tool.
        turn_modifier: Turn speed modifier to apply while the weapon is
            equipped.
        move_modifier: Move speed modifier to apply while the weapon is
            equipped.
        sprint_recovery_ms: Recovery time to allow firing or ADS after
            the player stopped sprinting.
        equip_ms: The weapon equip time in milliseconds.
        unequip_ms: The weapon unequip time in milliseconds.
        to_iron_sights_ms: The ADS enter time in milliseconds.
        from_iron_sights_ms: The ADS exit time in milliseconds.
        heat_capacity: The heat capacity of the weapon. Generally
            identical to :attr:`~auraxium.ps2.FireMode.heat_threshold`,
            but this only uses the first fire mode of the weapon.
        heat_bleed_off_rate: The rate at which the weapon will cool
            down after firing stops.
        heat_overheat_penalty_ms: The overheat penalty imposed if the
            user overheats the weapon.
        melee_detect_width: The hitbox width for melee weapons.
        melee_detect_height: The hitbox height for melee weapons.

    """

    weapon_id: int
    weapon_group_id: Optional[int] = None
    turn_modifier: float
    move_modifier: float
    sprint_recovery_ms: Optional[int] = None
    equip_ms: Optional[int]
    unequip_ms: Optional[int] = None
    to_iron_sights_ms: Optional[int] = None
    from_iron_sights_ms: Optional[int] = None
    heat_capacity: Optional[int] = None
    heat_bleed_off_rate: Optional[float] = None
    heat_overheat_penalty_ms: Optional[int] = None
    melee_detect_width: Optional[float] = None
    melee_detect_height: Optional[float] = None


class WeaponDatasheet(Ps2Data):
    """Data class for weapon datasheets.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    .. note::
        The data found herein is generally less reliable than the
        sources it was collected from, such as
        :class:`~auraxium.ps2.FireMode` or
        :class:`~auraxium.ps2.Projectile`.

        This data is mostly used for display in the depot and may not
        be inaccurate.

    Attributes:
        item_id: The item the datasheet is for.
        direct_damage: The direct damage amount.
        indirect_damage: The indirect damage amount.
        damage: (Not yet documented)
        damage_min: (Not yet documented)
        damage_max: (Not yet documented)
        fire_cone: (Not yet documented)
        fire_cone_min: (Not yet documented)
        fire_cone_max: (Not yet documented)
        fire_rate_ms: (Not yet documented)
        fire_rate_ms_min: (Not yet documented)
        fire_rate_mx_max: (Not yet documented)
        reload_ms: (Not yet documented)
        reload_ms_min: (Not yet documented)
        reload_ms_max: (Not yet documented)
        clip_size: The clip size for the default ammo type
        capacity: The total ammo capacity for the default ammo type
        range: The localised description of the weapon range (i.e.
            "Medium", "Long", etc.).
        show_clip_size: Whether to display the clip size data to the
            user.
        show_fire_modes: Whether to display the list of fire modes to
            the user.
        show_range: Whether to display the :attr:`range` data to the
            user.

    """

    item_id: int
    direct_damage: Optional[int] = None
    indirect_damage: Optional[int] = None
    damage: int
    damage_min: int
    damage_max: int
    fire_cone: float
    fire_cone_min: float
    fire_cone_max: float
    fire_rate_ms: int
    fire_rate_ms_min: int
    fire_rate_mx_max: int
    reload_ms: int
    reload_ms_min: int
    reload_ms_max: int
    clip_size: int
    capacity: int
    range: LocaleData
    show_clip_size: bool
    show_fire_modes: bool
    show_range: bool



"""Data classes for :mod:`auraxium.ps2.experience`."""

from ..base import Ps2Data
from ..types import LocaleData

__all__ = [
    'ExperienceData',
    'ExperienceRankData'
]


class ExperienceData(Ps2Data):
    """Data class for :class:`auraxium.ps2.experience.Experience`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        experience_id: The unique ID of this experience tick.
        description: A description of when this experience reward is
            granted.
        xp: The amount of experience points awarded.

    """

    experience_id: int
    description: str
    xp: int  # pylint: disable=invalid-name


class ExperienceRankData(Ps2Data):
    """Data class for :class:`auraxium.ps2.experience.Experience`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        rank: The battle rank value represented by this rank name.
        xp_max: (Not yet documented)
        vs: Empire-specific rank data for VS.
        vs_image_path: The VS-specific default image path.
        nc: Empire-specific rank data for NC.
        nc_image_path: The NC-specific default image path.
        tr: Empire-specific rank data for TR.
        tr_image_path: The TR-specific default image path.

    """

    class EmpireData(Ps2Data):
        """Object representation of an empire-specific sub-key.

        Attributes:
            title: The localised title for this experience rank.
            image_set_id: The image set associated with this rank.
            image_id: The default image asset for this rank.

        """

        title: LocaleData
        image_set_id: int
        image_id: int

    rank: int
    xp_max: int
    vs: EmpireData  # pylint: disable=invalid-name
    vs_image_path: str
    nc: EmpireData  # pylint: disable=invalid-name
    nc_image_path: str
    tr: EmpireData  # pylint: disable=invalid-name
    tr_image_path: str



"""Data classes for :mod:`auraxium.ps2.depot`."""

from typing import Optional

from ..base import Ps2Data
from ..types import LocaleData

__all__ = [
    'MarketingBundleData',
    'MarketingBundleSingleData'
]


class MarketingBundleData(Ps2Data):
    """Data class for :class:`auraxium.ps2.depot.MarketingBundle`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        marketing_bundle_id: The unique ID of this bundle.
        name: The localised name of the bundle.
        description: The description text for this bundle.
        image_id: The image asset ID for this bundle.
        cert_price: The unlock price in certification points, if any.
        station_cash_price: The unlock price in daybreak cash, if any.
        release_time: The time at which this bundle was first released
            as a UTC timestamp.

    """

    marketing_bundle_id: int
    name: LocaleData
    description: LocaleData
    image_id: int
    cert_price: Optional[int]
    station_cash_price: int
    release_time: int


class MarketingBundleSingleData(Ps2Data):
    """Data class for :class:`auraxium.ps2.depot.MarketingBundleSingle`.

    This is generally used for the single-item "bundles" in the depot,
    i.e. any single weapon or cosmetic purchasable on its own.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        marketing_bundle_id: The unique ID of this bundle.
        item_id: The item unlocked by this bundle.
        item_quantity: The number of items received.
        station_cash_price: The daybreak cash price of the item.
        cert_price: The certification point price of the item.
        release_time: The time at which this item was first released
            as a UTC timestamp.

    """

    marketing_bundle_id: int
    item_id: int
    item_quantity: int
    station_cash_price: int
    cert_price: Optional[int]
    release_time: int



"""Data classes for :mod:`auraxium.ps2.fire`."""

from typing import Optional

from ..base import Ps2Data
from ..types import LocaleData

__all__ = [
    'FireModeData',
    'FireGroupData'
]


class FireModeData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.FireMode`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        fire_mode_id: The unique ID of this fire mode.
        fire_mode_type_id: The type of fire mode as a value of the
            :class:`FireModeType` enumerator.
        ability_id: (Not yet documented)
        ammo_slot: The ammo slot used by this fire mode.
        automatic: Whether the fire mode is automatic or not.
        grief_immune: Whether this fire mode is excluded from friendly
            fire.
        iron_sights: Whether this weapon comes with iron sights.
        laser_guided: Whether this weapon is laser guided.
        move_modifier: (Not yet documented)
        projectile_speed_override: A fire mode specific override used
            to override the fire mode's projectile's speed.
        sprint_fire: Whether the fire mode allows firing and sprinting.
        turn_modifier: (Not yet documented)
        use_in_water: Unused.
        zoom_default: The iron-sight zoom level of the fire mode.
        cof_override: (Not yet documented)
        cof_pellet_spread: The pellet spread used for multi-projectile
            weapons (such as non-slug shotguns or the NSX Tengu).
        cof_range: The distance the projectile may travel before
            being destroyed.
        cof_recoil: The cone-of-fire bloom per shot.
        cof_scalar: The starting cone-of-fire scalar. This exists so it
            can be modified by attachments.
        cof_scalar_moving: The starting cone-of-fire scalar
            while moving. This exists so it can be modified by
            attachments.
        player_state_group_id: The state group associated with this
            fire mode. State groups contain the state-dependent values
            for a weapon, such as accuracy changes from sprinting,
            crouching, etc.
        damage_direct_effect_id: The damage effect used on direct hits.
        damage_head_multiplier: The headshot multiplier for the fire
            mode.
        damage_indirect_effect_id: The splash damage effect used on
            indirect hits.
        damage_legs_multiplier: The leg shot multiplier for the fire
            mode.
        fire_ammo_per_shot: The amount of ammo consumed per shot. Some
            weapons, like the VS Lancer, may consume more than 1 bullet
            per shot.
        fire_auto_fire_ms: The time in-between shots during automatic
            fire.
        fire_burst_count: The burst size of the fire mode.
        fire_charge_up_ms: The time it takes the weapon to charge up,
            used in the VS Lancer rocket launcher.
        fire_delay_ms: The delay between pulling the trigger and
            the weapon actually firing, used in the NXS Yumi.
        fire_detect_range: (Not yet documented)
        fire_duration_ms: The firing duration of a weapon. Used for
            grenade throwing and knifing animations.
        fire_refire_ms: The re-fire delay used for semi-automatic
            weapons.
        fire_pellets_per_shot: The number of pellets fired by non-slug
            shotguns or the NSX Tengu.
        heat_per_shot: The weapon heat applied per shot.
        heat_recovery_delay_ms: The duration after which weapon heat
            will start to dissipate.
        heat_threshold: The overheating threshold of the weapon. This
            is comparable to the magazine size for heat-based weapons,
            with the number of shots available without overheating
            being equal to the floor of
            :attr:`heat_threshold` divided :attr:`heat_per_shot`.
        lockon_acquire_close_ms: The minimum distance from the target
            to establish a lock-on.
        lockon_acquire_far_ms: The maximum distance from the target to
            establish a lock-on.
        lockon_acquire_ms: The lock-on time.
        lockon_angle: The maximum error allowed while locking on, you
            can think of this like a cone of fire.
        lockon_lose_ms: (Not yet documented)
        lockon_maintain: (Not yet documented)
        lockon_radius: Not used as of June 2020, use
            :attr:`lockon_range` instead.
        lockon_range: The maximum lock-on range.
        lockon_range_close: (Not yet documented)
        lockon_range_far: (Not yet documented)
        lockon_required: Whether this weapon can be fired without
            being locked on (True for the NS Annihilator, False for
            faction-specific lock-on launchers).
        recoil_angle_max: The maximum recoil angle, used to randomise
            recoil angle.
        recoil_angle_min: The minimum recoil angle, used to randomise
            recoil angle.
        recoil_first_shot_modifier: The first shot multiplier for the
            weapon.
        recoil_horizontal_max: The maximum horizontal recoil per shot.
        recoil_horizontal_max_increase: The maximum horizontal recoil
            jump per shot.
        recoil_horizontal_min: The minimum horizontal recoil per shot.
        recoil_horizontal_min_increase: The minimum horizontal recoil
            jump per shot.
        recoil_horizontal_tolerance: (Not yet documented)
        recoil_increase: The base vertical recoil while standing.
        recoil_increase_crouched: The base vertical recoil while
            crouching.
        recoil_magnitude_max: The maximum vertical recoil per shot.
        recoil_magnitude_min: The minimum vertical recoil per shot.
        recoil_max_total_magnitude: The total maximum recoil, including
            both horizontal and vertical offsets.
        recoil_recovery_acceleration: (Not yet documented)
        recoil_recovery_delay_ms: The delay before the weapon will
            reset after the player stopped firing.
        recoil_recovery_rate: The speed at which the weapon will reset
            after the players topped firing.
        recoil_shots_at_min_magnitude: (Not yet documented)
        reload_block_auto: (Not yet documented)
        reload_continuous: (Unused, always False or not set.)
        reload_ammo_fill_ms: The time between ammo refill ticks.
        reload_chamber_ms: The chamber time for bolt-action weapons.
        reload_loop_start_ms: The initial loop time for reloading with
            looped reload weapons (such as pump action shotguns).
        reload_loop_end_ms: The final delay after reloading a looped
            reload weapon is completed.
        reload_time_ms: The reload time of the weapon.
        sway_amplitude_x: Sway amplitude in X direction.
        sway_amplitude_y: Sway amplitude in Y direction.
        sway_can_steady: Whether the player can hold their breath to
            stop camera sway.
        sway_period_x: Sway period in X direction.
        sway_period_y: Sway period in Y direction.
        armor_penetration: (Unused as of June 2020)
        max_damage: Maximum direct damage of the weapon.
        max_damage_ind: Maximum indirect damage of the weapon.
        max_damage_ind_radius: The radius over which maximum indirect
            damage of the weapon will fall off.
        max_damage_range: The range up to which the weapon will deal
            its maximum damage.
        min_damage: The minimum direct damage of the weapon.
        min_damage_ind: The minimum indirect damage of the weapon.
        min_damage_ind_radius: The radius over which the minimum
            indirect damage of the weapon will fall off.
        min_damage_range: The range at which the weapon deals its
            minimum damage.
        shield_bypass_pct: The percentage of damage that will bypass
            shields and always affect the raw health pool.
        description: Localised description of the fire mode (e.g.
            "Automatic").

    """

    fire_mode_id: int
    fire_mode_type_id: int
    ability_id: Optional[int] = None
    ammo_slot: Optional[int] = None
    automatic: bool
    grief_immune: Optional[bool] = None
    iron_sights: Optional[bool] = None
    laser_guided: Optional[bool] = None
    move_modifier: float
    projectile_speed_override: Optional[int] = None
    sprint_fire: Optional[bool] = None
    turn_modifier: float
    use_in_water: Optional[bool] = None
    zoom_default: float
    cof_override: Optional[float] = None
    cof_pellet_spread: Optional[float] = None
    cof_range: float
    cof_recoil: Optional[float] = None
    cof_scalar: float
    cof_scalar_moving: float
    player_state_group_id: int
    damage_direct_effect_id: Optional[int] = None
    damage_head_multiplier: Optional[float] = None
    damage_indirect_effect_id: Optional[int] = None
    damage_legs_multiplier: Optional[float] = None
    fire_ammo_per_shot: int
    fire_auto_fire_ms: Optional[int] = None
    fire_burst_count: int
    fire_charge_up_ms: Optional[int] = None
    fire_delay_ms: Optional[int] = None
    fire_detect_range: float
    fire_duration_ms: Optional[int] = None
    fire_refire_ms: int
    fire_pellets_per_shot: Optional[int] = None
    heat_per_shot: Optional[int] = None
    heat_recovery_delay_ms: Optional[int] = None
    heat_threshold: Optional[int] = None
    lockon_acquire_close_ms: Optional[int] = None
    lockon_acquire_far_ms: Optional[int] = None
    lockon_acquire_ms: Optional[int] = None
    lockon_angle: Optional[float] = None
    lockon_lose_ms: Optional[int] = None
    lockon_maintain: Optional[bool] = None
    lockon_radius: Optional[float] = None  # Not used as of June 2020
    lockon_range: Optional[float] = None  # Maybe this replaced "lockon_radius"
    lockon_range_close: Optional[float] = None
    lockon_range_far: Optional[float] = None
    lockon_required: Optional[bool] = None
    recoil_angle_max: Optional[float] = None
    recoil_angle_min: Optional[float] = None
    recoil_first_shot_modifier: float
    recoil_horizontal_max: Optional[float] = None
    recoil_horizontal_max_increase: Optional[float] = None
    recoil_horizontal_min: Optional[float] = None
    recoil_horizontal_min_increase: Optional[float] = None
    recoil_horizontal_tolerance: Optional[float] = None
    recoil_increase: Optional[float] = None
    recoil_increase_crouched: Optional[float] = None
    recoil_magnitude_max: Optional[float] = None
    recoil_magnitude_min: Optional[float] = None
    recoil_max_total_magnitude: Optional[float] = None
    recoil_recovery_acceleration: Optional[int] = None
    recoil_recovery_delay_ms: Optional[int] = None
    recoil_recovery_rate: Optional[int] = None
    recoil_shots_at_min_magnitude: Optional[bool] = None
    reload_block_auto: Optional[bool] = None
    reload_continuous: Optional[bool] = None  # Always False or None
    reload_ammo_fill_ms: Optional[int] = None
    reload_chamber_ms: Optional[int] = None
    reload_loop_start_ms: Optional[int] = None
    reload_loop_end_ms: Optional[int] = None
    reload_time_ms: int
    sway_amplitude_x: Optional[float] = None
    sway_amplitude_y: Optional[float] = None
    sway_can_steady: Optional[bool] = None
    sway_period_x: Optional[int] = None
    sway_period_y: Optional[int] = None
    armor_penetration: Optional[float] = None  # Always zero or None
    max_damage: Optional[int] = None
    max_damage_ind: Optional[int] = None
    max_damage_ind_radius: Optional[float] = None
    max_damage_range: Optional[float] = None
    min_damage: Optional[int] = None
    min_damage_ind: Optional[int] = None
    min_damage_ind_radius: Optional[float] = None
    min_damage_range: Optional[float] = None
    shield_bypass_pct: Optional[int] = None
    description: LocaleData


class FireGroupData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.FireGroup`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        fire_group_id: The unique ID of this fire group.
        chamber_duration_ms: The rechamber time for weapons in this
            group.
        transition_duration_ms: (Not yet documented)
        spool_up_ms: The duration of the spool-up period for this
            weapon group.
        spool_up_initial_refire_ms: The initial fire speed (rounds
            per minute) of the weapon group. The weapon starts out at
            this value when firing, then tapers to the regular value
            after :attr:`spool_up_ms` milliseconds.
        can_chamber_ironsights: Whether a bolt-action weapon can be
            rechambered while in ADS.

    """

    fire_group_id: int
    chamber_duration_ms: Optional[int] = None
    transition_duration_ms: Optional[int] = None
    spool_up_ms: Optional[int] = None
    spool_up_initial_refire_ms: Optional[int] = None
    can_chamber_ironsights: Optional[bool] = None



"""Player state group class definitions."""

import enum
from ..models import PlayerStateGroup

__all__ = [
    'PlayerState',
    'PlayerStateGroup'
]


class PlayerState(enum.IntEnum):
    """A player state.

    These are used to apply the situational weapon stats depending on
    the player state, e.g. running, springing, crouching, etc.
    """

    STANDING = 0
    CROUCHING = 1
    STANDING_MOVING = 2
    SPRINTING = 3
    FALLING_LONG = 4
    CROUCHING_MOVING = 5



"""Item and item attachment class definitions."""

from typing import Final, TYPE_CHECKING

from ..base import Cached, ImageMixin, Named
from ..census import Query
from ..models import ItemCategoryData, ItemData, ItemTypeData
from ..request import extract_single
from ..proxy import InstanceProxy, SequenceProxy

from .faction import Faction
from .profile import Profile

if TYPE_CHECKING:
    # This is only imported during static type checking to resolve the forward
    # references. This avoids a circular import at runtime.
    from .weapon import Weapon, WeaponDatasheet


class ItemCategory(Named, cache_size=32, cache_ttu=3600.0):
    """A category of item.

    This represents the item filter views used in the in-game depot,
    such as "Infantry Gear", "Weapon Camo" or "Vehicle Weapons".
    """

    collection = 'item_category'
    data: ItemCategoryData
    dataclass = ItemCategoryData
    id_field = 'item_category_id'


class ItemType(Cached, cache_size=10, cache_ttu=60.0):
    """A type of item.

    Item types are a coarser classification used internally. They are
    used to differentiate between tangible items like weapons or
    consumables, cosmetics, implants, as well as abstract item-like
    utilities like loadout slots, server transfers, or name change
    tokens.
    """

    collection = 'item_type'
    data: ItemTypeData
    dataclass = ItemTypeData
    id_field = 'item_type_id'


class Item(Named, ImageMixin, cache_size=128, cache_ttu=3600.0):
    """An item that may be owned by a character.

    This includes the item component of weapons, which is extended by
    weapon specific data in the associated
    :class:`auraxium.ps2.weapon.Weapon` instance.
    """

    collection = 'item'
    data: ItemData
    dataclass = ItemData
    id_field = 'item_id'

    def attachments(self) -> SequenceProxy['Item']:
        """Return the attachment options for this item.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'item_attachment'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(100)
        join = query.create_join(self.collection)
        join.set_fields('attachment_item_id', self.id_field)
        return SequenceProxy(self.__class__, query, client=self._client)

    def category(self) -> InstanceProxy[ItemCategory]:
        """Return the category of the item.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        if self.data.item_category_id is None:
            raise ValueError(f'{self} does not define a category')
        query = Query(
            ItemCategory.collection, service_id=self._client.service_id)
        query.add_term(
            field=ItemCategory.id_field, value=self.data.item_category_id)
        return InstanceProxy(ItemCategory, query, client=self._client)

    def description(self, locale: str = 'en') -> str:
        """Return the description of this item in the given locale.

        This will return "Missing String" if no string has been set for
        the selected locale.
        """
        if hasattr(self.data.description, locale):
            return str(getattr(self.data.description, locale))
        return 'Missing String'

    def faction(self) -> InstanceProxy[Faction]:
        """Return the faction that has access to this item.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        value = self.data.faction_id or -1
        query = Query(Faction.collection, service_id=self._client.service_id)
        query.add_term(field=Faction.id_field, value=value)
        return InstanceProxy(Faction, query, client=self._client)

    async def datasheet(self) -> 'WeaponDatasheet':
        """Return the datasheet for the weapon."""
        # pylint: disable=import-outside-toplevel
        from .weapon import WeaponDatasheet
        collection: Final[str] = 'weapon_datasheet'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        payload = await self._client.request(query)
        data = extract_single(payload, collection)
        return WeaponDatasheet(**data)

    def profiles(self) -> SequenceProxy[Profile]:
        """Return the profiles the item is available to.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'item_profile'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(50)
        join = query.create_join(Profile.collection)
        join.set_fields(Profile.id_field)
        return SequenceProxy(Profile, query, client=self._client)

    def type(self) -> InstanceProxy[ItemType]:
        """Return the type of item.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        if self.data.item_type_id is None:
            raise ValueError(f'{self} does not define a type')
        query = Query(ItemType.collection, service_id=self._client.service_id)
        query.add_term(field=ItemType.id_field, value=self.data.item_type_id)
        return InstanceProxy(ItemType, query, client=self._client)

    def weapon(self) -> InstanceProxy['Weapon']:
        """Return the weapon associated with this item, if any.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        from .weapon import Weapon  # pylint: disable=import-outside-toplevel
        collection: Final[str] = 'item_to_weapon'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        join = query.create_join('weapon')
        join.set_fields('weapon_id')
        return InstanceProxy(Weapon, query, client=self._client)



"""Object definition for the faction type."""

from ..base import ImageMixin, Named
from ..models import FactionData


class Faction(Named, ImageMixin, cache_size=10):
    """A faction in PS2."""

    collection = 'faction'
    data: FactionData
    dataclass = FactionData
    id_field = 'faction_id'

    def __repr__(self) -> str:
        """Return the unique string representation of the faction.

        This will take the form of <class:id:tag>, e.g. <Faction:2:NC>.

        Returns:
            A string representing the object.

        """
        return f'<{self.__class__.__name__}:{self.id}:{self.data.code_tag}>'

    @property
    def tag(self) -> str:
        """Return the tag of this faction (VS, TR, NC, or NSO)."""
        return str(self.data.code_tag)



"""Skill and skill line class definitions."""

from ..base import ImageMixin, Named
from ..census import Query
from ..models import SkillData, SkillCategoryData, SkillLineData, SkillSetData
from ..proxy import InstanceProxy, SequenceProxy

from .item import Item


class SkillSet(Named, ImageMixin, cache_size=100, cache_ttu=60.0):
    """A skill set for a particular vehicle or class."""

    collection = 'skill_set'
    data: SkillSetData
    dataclass = SkillSetData
    id_field = 'skill_set_id'

    def categories(self) -> SequenceProxy['SkillCategory']:
        """Return the skill categories in this skill set.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        query = Query(
            SkillCategory.collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(100)
        return SequenceProxy(SkillCategory, query, client=self._client)

    def required_item(self) -> InstanceProxy[Item]:
        """Return the item required for access to this skill set.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        item_id = self.data.required_item_id or -1
        query = Query(Item.collection, service_id=self._client.service_id)
        query.add_term(field=Item.id_field, value=item_id)
        return InstanceProxy(Item, query, client=self._client)


class SkillCategory(Named, ImageMixin, cache_size=50, cache_ttu=60.0):
    """A skill category for a particular class or vehicle.

    Skill categories are groups like "Passive Systems" or "Performance
    Slot".
    """

    collection = 'skill_category'
    data: SkillCategoryData
    dataclass = SkillCategoryData
    id_field = 'skill_category_id'

    def skill_lines(self) -> SequenceProxy['SkillLine']:
        """Return the skill lines contained in this category.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        query = Query(SkillLine.collection, service_id=self._client.service_id)
        query.add_term(field=SkillCategory.id_field, value=self.id)
        query.sort('skill_category_index')
        return SequenceProxy(SkillLine, query, client=self._client)

    def skill_set(self) -> InstanceProxy['SkillSet']:
        """Return the skill set for this category.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(SkillSet.collection, service_id=self._client.service_id)
        query.add_term(field=SkillSet.id_field, value=self.data.skill_set_id)
        return InstanceProxy(SkillSet, query, client=self._client)


class SkillLine(Named, ImageMixin, cache_size=50, cache_ttu=60.0):
    """A series of skills or certifications."""

    collection = 'skill_line'
    data: SkillLineData
    dataclass = SkillLineData
    id_field = 'skill_line_id'

    def category(self) -> InstanceProxy[SkillCategory]:
        """Return the category this skill line belongs to.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        category_id = self.data.skill_category_id or -1
        query = Query(
            SkillCategory.collection, service_id=self._client.service_id)
        query.add_term(field=SkillCategory.id_field, value=category_id)
        return InstanceProxy(SkillCategory, query, client=self._client)

    def skills(self) -> SequenceProxy['Skill']:
        """Return the skills contained in this skill line in order.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        query = Query(Skill.collection, service_id=self._client.service_id)
        query.add_term(field=SkillLine.id_field, value=self.id)
        query.limit(20)
        query.sort('skill_line_index')
        return SequenceProxy(Skill, query, client=self._client)


class Skill(Named, ImageMixin, cache_size=50, cache_ttu=60.0):
    """A skill or certification unlockable by a character."""

    collection = 'skill'
    data: SkillData
    dataclass = SkillData
    id_field = 'skill_id'

    def grant_item(self) -> InstanceProxy[Item]:
        """Return the item unlocked by this skill.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        item_id = self.data.grant_item_id or -1
        query = Query(Item.collection, service_id=self._client.service_id)
        query.add_term(field=Item.id_field, value=item_id)
        return InstanceProxy(Item, query, self._client)

    def skill_line(self) -> InstanceProxy[SkillLine]:
        """Return the skill line containing this skill.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(SkillLine.collection, service_id=self._client.service_id)
        query.add_term(field=SkillLine.id_field, value=self.data.skill_line_id)
        return InstanceProxy(SkillLine, query, self._client)



"""Data classes for :mod:`auraxium.ps2.effect`."""

from typing import Optional

from ..base import Ps2Data

__all__ = [
    'EffectData',
    'EffectTypeData'
]


class EffectData(Ps2Data):
    """Data class for :class:`auraxium.ps2.effect.Effect`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        effect_id: The unique ID of this effect.
        effect_type_id: The associated effect type for this effect.
        ability_id: The ability spawning the effect, if any.
        target_type_id: Integer value of the :class:`TargetType`
            enumerator used to find targets for this effect.
        resist_type_id: The :class:`~auraxium.ps2.ResistInfo` entry
            used by this effect.
        is_drain: (Not yet documented)
        duration_seconds: The duration of the effect.
        param*: Type-specific parameters for this effect. Refer to the
            corresponding :class:`EffectType` for details.

    """

    effect_id: int
    effect_type_id: int
    ability_id: Optional[int]
    target_type_id: Optional[int]
    resist_type_id: int
    is_drain: Optional[bool]
    duration_seconds: Optional[float]
    param1: Optional[str]
    param2: Optional[str]
    param3: Optional[str]
    param4: Optional[str]
    param5: Optional[str]
    param6: Optional[str]
    param7: Optional[str]
    param8: Optional[str]
    param9: Optional[str]
    param10: Optional[str]
    param11: Optional[str]
    param12: Optional[str]
    param13: Optional[str]


class EffectTypeData(Ps2Data):
    """Data class for :class:`auraxium.ps2.effect.EffectType`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        effect_type_id: The unique ID of this effect type.
        description: A description of what this effect type is used
            for.
        param*: Descriptions of what the corresponding parameter is
            used for in abilities of this type.

    """

    effect_type_id: int
    description: str
    param1: Optional[str]
    param2: Optional[str]
    param3: Optional[str]
    param4: Optional[str]
    param5: Optional[str]
    param6: Optional[str]
    param7: Optional[str]
    param8: Optional[str]
    param9: Optional[str]
    param10: Optional[str]
    param11: Optional[str]
    param12: Optional[str]
    param13: Optional[str]



"""Outfit and outfit member class definitions."""

import logging
from typing import ClassVar, Final, List, Optional, TYPE_CHECKING, Union

from ..base import Cached, Named
from ..cache import TLRUCache
from ..census import Query
from ..client import Client
from ..errors import NotFoundError
from ..models import OutfitData, OutfitMemberData, OutfitRankData
from ..proxy import InstanceProxy, SequenceProxy
from ..request import extract_payload, extract_single

if TYPE_CHECKING:
    # This is only imported during static type checking to resolve the
    # 'Character' forward reference. This avoids a circular import at runtime.
    from .character import Character

log = logging.getLogger('auraxium.ps2')


class OutfitMember(Cached, cache_size=100, cache_ttu=300.0):
    """A member of an outfit.

    This class can be treated as an extension of the
    :class:`auraxium.ps2.character.Character` class.
    """

    collection = 'outfit_member'
    data: OutfitMemberData
    dataclass = OutfitMemberData
    id_field = 'character_id'

    def character(self) -> InstanceProxy['Character']:
        """Return the character associated with this member.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        # NOTE: This is required due to OutfitMember effectively being an
        # extension of Character.
        # pylint: disable=import-outside-toplevel
        from .character import Character
        query = Query(Character.collection, service_id=self._client.service_id)
        query.add_term(field=Character.id_field, value=self.data.character_id)
        return InstanceProxy(Character, query, client=self._client)

    def outfit(self) -> InstanceProxy['Outfit']:
        """Return the character associated with this member.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Outfit.collection, service_id=self._client.service_id)
        query.add_term(field=Outfit.id_field, value=self.data.outfit_id)
        return InstanceProxy(Outfit, query, client=self._client)


class Outfit(Named, cache_size=20, cache_ttu=300.0):
    """A player-run outfit."""

    _cache: ClassVar[TLRUCache[Union[int, str], 'Outfit']]
    collection = 'outfit'
    data: OutfitData
    dataclass = OutfitData
    id_field = 'outfit_id'

    @classmethod
    async def get_by_name(cls, name: str, *, locale: str = 'en', client: Client
                          ) -> Optional['Outfit']:
        """Retrieve an outfit by its unique name.

        This query is always case-insensitive.

        """
        log.debug('%s "%s"[%s] requested', cls.__name__, name, locale)
        if (instance := cls._cache.get(f'_{name.lower()}')) is not None:
            log.debug('%r restored from cache', instance)
            return instance
        log.debug('%s "%s"[%s] not cached, generating API query...',
                  cls.__name__, name, locale)
        query = Query(cls.collection, service_id=client.service_id,
                      name_lower=name.lower()).limit(1)
        data = await client.request(query)
        try:
            payload = extract_single(data, cls.collection)
        except NotFoundError:
            return None
        return cls(payload, client=client)

    @classmethod
    async def get_by_tag(cls, tag: str, client: Client) -> Optional['Outfit']:
        """Return an outfit by its unique tag.

        This query is always case-insensitive.

        """
        log.debug('%s with tag "%s" requested, generating API query...',
                  cls.__name__, tag)
        query = Query(cls.collection, service_id=client.service_id,
                      alias_lower=tag.lower()).limit(1)
        data = await client.request(query)
        try:
            payload = extract_single(data, cls.collection)
        except NotFoundError:
            return None
        return cls(payload, client=client)

    def leader(self) -> InstanceProxy[OutfitMember]:
        """Return the current leader of the outfit.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(
            OutfitMember.collection, service_id=self._client.service_id)
        query.add_term(
            field=OutfitMember.id_field, value=self.data.leader_character_id)
        return InstanceProxy(OutfitMember, query, client=self._client)

    def members(self) -> SequenceProxy[OutfitMember]:
        """Return the members of the outfit.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        query = Query(
            OutfitMember.collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(5000)
        return SequenceProxy(OutfitMember, query, client=self._client)

    def name(self, locale: str = 'en') -> str:
        """Return the unique name of the outfit.

        Since outfit names are not localised, the "locale" keyword
        argument is ignored.

        This will always return the capitalised version of the name.
        Use the built-int str.lower() method for a lowercase version.
        """
        return str(self.data.name)

    async def ranks(self) -> List[OutfitRankData]:
        """Return the list of ranks for the outfit."""
        collection: Final[str] = 'outfit_rank'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(20)
        data = await self._client.request(query)
        payload = extract_payload(data, collection)
        return [OutfitRankData(**c) for c in payload]



"""Data classes for :mod:`auraxium.ps2.achievement`."""

from ..base import ImageData, Ps2Data
from ..types import LocaleData

__all__ = [
    'AchievementData'
]


class AchievementData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.achievement.Achievement`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        achievement_id: The unique ID of this achievement.
        item_id: The item associated with this achievement. An item ID
            of ``0`` signifies that this achievement is a ribbon not
            tied to any weapon.
        objective_group_id: The objective group tied to this
            achievement.
        reward_id: The reward granted when this achievement is earned.
        repeatable: Whether this achievement is repeatable. Ribbons
            generally are repeatable, weapon medals are not.
        name: The localised name of the achievement.
        description: The localised description of achievement.

    """

    achievement_id: int
    item_id: int
    objective_group_id: int
    reward_id: int
    repeatable: bool
    name: LocaleData
    description: LocaleData



"""Data classes for :mod:`auraxium.ps2.directive`."""

from typing import Optional

from ..base import ImageData, Ps2Data
from ..types import LocaleData

__all__ = [
    'DirectiveData',
    'DirectiveTierData',
    'DirectiveTreeData',
    'DirectiveTreeCategoryData'
]


class DirectiveData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.directive.Directive`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        directive_id: The unique ID of this directive.
        directive_tree_id: The directive tree of this directive.
        directive_tier_id: The directive tier of this directive.
        objective_set_id: The objective set contributing towards this
            directive.
        qualify_requirement_id: An item that must be unlocked for this
            directive to be available.
        name: The localised name of the directive.
        name: The localised description of the directive.


    """

    directive_id: int
    directive_tree_id: int
    directive_tier_id: int
    objective_set_id: int
    qualify_requirement_id: Optional[int] = None
    name: LocaleData
    description: Optional[LocaleData] = None


class DirectiveTierData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.directive.DirectiveTier`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        directive_tier_id: The unique ID of the directive tier.
        directive_tree_id: The directive tree this directive belongs
            to.
        reward_set_id: The reward set awarded upon completion of this
            directive tier.
        directive_points: The directive points awarded upon completion
            of this directive tier.
        completion_count: The number of directives that must be
            completed for completion of this directive tier.
        name: The localised name of the directive tier.

    """

    directive_tier_id: int
    directive_tree_id: int
    reward_set_id: Optional[int] = None
    directive_points: int
    completion_count: int
    name: LocaleData


class DirectiveTreeData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.directive.DirectiveTree`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        directive_tree_id: The unique ID of the directive tree.
        directive_tree_category_id: The category of the directive tree.
        name: The localised name of the directive tree.
        description: The localised description of the directive tree.

    """

    directive_tree_id: int
    directive_tree_category_id: int
    name: LocaleData
    description: Optional[LocaleData] = None


class DirectiveTreeCategoryData(Ps2Data):
    """Data class for :class:`auraxium.ps2.directive.DirectiveTreeCategory`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        directive_tree_category_id: The unique ID of the directive tree
            category.
        name: The localised name of the directive tree category.

    """

    directive_tree_category_id: int
    name: LocaleData



"""Character class definition."""

import logging
from typing import Any, ClassVar, Final, List, Optional, Tuple, Union

from ..base import Named
from ..cache import TLRUCache
from ..census import Query
from ..client import Client
from ..errors import NotFoundError
from ..models import CharacterAchievement, TitleData, CharacterData
from ..proxy import InstanceProxy, SequenceProxy
from ..request import extract_payload, extract_single
from ..types import CensusData

from .faction import Faction
from .item import Item
from .outfit import Outfit, OutfitMember
from .profile import Profile
from .world import World

log = logging.getLogger('auraxium.ps2')


class Title(Named, cache_size=300, cache_ttu=300.0):
    """A title selectable by a character."""

    collection = 'title'
    data: TitleData
    dataclass = TitleData
    id_field = 'title_id'


class Character(Named, cache_size=256, cache_ttu=30.0):
    """A player-controlled fighter."""

    _cache: ClassVar[TLRUCache[Union[int, str], 'Character']]
    collection = 'character'
    data: CharacterData
    dataclass = CharacterData
    id_field = 'character_id'

    async def achievements(self, **kwargs: Any) -> List[CharacterAchievement]:
        """Return the achievement status for a character."""
        collection: Final[str] = 'characters_achievement'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.limit(5000)
        query.add_term(field=self.id_field, value=self.id)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return [CharacterAchievement(**d) for d in data]

    async def currency(self) -> Tuple[int, int]:
        """Return the currencies of the character.

        This returns a tuple of the number of Nanites and ASP tokens
        available to the player.

        Other currencies are not exposed to the API as of the writing
        of this module.
        """
        collection: Final[str] = 'characters_currency'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        payload = await self._client.request(query)
        data = extract_single(payload, collection)
        return int(data['quantity']), int(data['prestige_currency'])

    async def directive(self, results: int = 1,
                        **kwargs: Any) -> List[CensusData]:
        """Query the directive status for this character."""
        collection: Final[str] = 'characters_directive'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(results)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return data

    async def directive_objective(self, results: int = 1,
                                  **kwargs: Any) -> List[CensusData]:
        """Query the objective status for a directive."""
        collection: Final[str] = 'characters_directive_objective'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(results)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return data

    async def directive_tier(self, results: int = 1,
                             **kwargs: Any) -> List[CensusData]:
        """Query the directive tier status for this character."""
        collection: Final[str] = 'characters_directive_tier'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(results)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return data

    async def directive_tree(self, results: int = 1,
                             **kwargs: Any) -> List[CensusData]:
        """Query the directive tree status for this character."""
        collection: Final[str] = 'characters_directive_tree'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(results)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return data

    async def events(self, **kwargs: Any) -> List[CensusData]:
        """Return and process past events for this character.

        This provides a REST endpoint for past character events.

        This is always limited to at most 1000 return values. Use the
        begin and end parameters to poll more data.
        """
        collection: Final[str] = 'characters_event'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(1000)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection=collection)
        return data

    async def events_grouped(self, **kwargs: Any) -> List[CensusData]:
        """Used to obtain kills and deaths on a per-player basis."""
        collection: Final[str] = 'characters_event_grouped'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(100_000)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection=collection)
        return data

    def faction(self) -> InstanceProxy[Faction]:
        """Return the faction of the character.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Faction.collection, service_id=self._client.service_id)
        query.add_term(field=Faction.id_field, value=self.data.faction_id)
        return InstanceProxy(Faction, query, client=self._client)

    async def friends(self) -> List['Character']:
        """Return the friends list of the character."""
        # NOTE: Due to the strange formatting of the characters_friend
        # collection, I does not seem possible to retrieve the associated
        # characters through joins.
        # This is solved through a second query matching the IDs, though this
        # does slow this query down dramatically.
        collection: Final[str] = 'characters_friend'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=Character.id_field, value=self.id)
        join = query.create_join(self.collection)
        join.set_list(True)
        payload = await self._client.request(query)
        data = extract_single(payload, collection)
        character_ids: List[str] = [
            str(d['character_id']) for d in data['friend_list']]
        characters = await Character.find(
            results=len(character_ids), client=self._client,
            character_id=','.join(character_ids))
        return characters

    @classmethod
    async def get_by_name(cls, name: str, *, locale: str = 'en', client: Client
                          ) -> Optional['Character']:
        """Retrieve an object by its unique name.

        This query is always case-insensitive.
        """
        log.debug('%s "%s"[%s] requested', cls.__name__, name, locale)
        if (instance := cls._cache.get(f'_{name.lower()}')) is not None:
            log.debug('%r restored from cache', instance)
            return instance
        log.debug('%s "%s"[%s] not cached, generating API query...',
                  cls.__name__, name, locale)
        query = Query(cls.collection, service_id=client.service_id,
                      name__first_lower=name.lower()).limit(1)
        data = await client.request(query)
        try:
            payload = extract_single(data, cls.collection)
        except NotFoundError:
            return None
        return cls(payload, client=client)

    @classmethod
    async def get_online(cls, id_: int, *args: int, client: Client
                         ) -> List['Character']:
        """Retrieve the characters that are online from a list."""
        char_ids = [id_]
        char_ids.extend(args)
        log.debug('Retrieving online status for %s characters', len(char_ids))
        query = Query(cls.collection, service_id=client.service_id,
                      character_id=','.join(str(c) for c in char_ids))
        query.limit(len(char_ids)).resolve('online_status')
        data = await client.request(query)
        payload = extract_payload(data, cls.collection)
        return [cls(c, client=client) for c in payload
                if int(c['online_status'])]

    def items(self) -> SequenceProxy[Item]:
        """Return the items available to the character.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'characters_item'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(5000)
        join = query.create_join(Item.collection)
        join.set_fields(Item.id_field)
        return SequenceProxy(Item, query, client=self._client)

    async def is_online(self) -> bool:
        """Return whether the given character is online."""
        return bool(int(await self.online_status()))

    def name(self, locale: str = 'en') -> str:
        """Return the unique name of the player.

        Since character names are not localised, the "locale" keyword
        argument is ignored.

        This will always return the capitalised version of the name.
        Use the built-int str.lower() method for a lowercase version.
        """
        return str(self.data.name.first)

    async def name_long(self) -> str:
        """Return the full name of the player.

        This includes an optional player title if the player has
        selected one.
        """
        if (title := await self.title()) is not None:
            return f'{title.name()} {self.name()}'
        return self.name()

    async def online_status(self) -> int:
        """Return the online status of the character.

        This returns 0 if the character is offline, or the world_id of
        the server they are logged into.
        """
        collection: Final[str] = 'characters_online_status'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        payload = await self._client.request(query)
        data = extract_single(payload, collection)
        return int(data['online_status'])

    def outfit(self) -> InstanceProxy[Outfit]:
        """Return the outfit of the character, if any.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        collection: Final[str] = 'outfit_member_extended'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        return InstanceProxy(Outfit, query, client=self._client)

    def outfit_member(self) -> InstanceProxy[OutfitMember]:
        """Return the outfit member of the character, if any.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(
            OutfitMember.collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        return InstanceProxy(OutfitMember, query, client=self._client)

    def profile(self) -> InstanceProxy[Profile]:
        """Return the last played profile of the character.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Profile.collection, service_id=self._client.service_id)
        query.add_term(field=Profile.id_field, value=self.data.profile_id)
        return InstanceProxy(Profile, query, client=self._client)

    async def skill(self, results: int = 1, **kwargs: Any) -> List[CensusData]:
        """Return the skills unlocked by this character."""
        collection: Final[str] = 'characters_skill'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(results)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return data

    async def stat(self, results: int = 1, **kwargs: Any) -> List[CensusData]:
        """Return global statistics for this character."""
        collection: Final[str] = 'characters_stat'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(results)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return data

    async def stat_by_faction(self, results: int = 1,
                              **kwargs: Any) -> List[CensusData]:
        """Return faction-specific statistics for this character."""
        collection: Final[str] = 'characters_stat_by_faction'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(results)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return data

    async def stat_history(self, results: int = 1,
                           **kwargs: Any) -> List[CensusData]:
        """Return historical statistics for this character."""
        collection: Final[str] = 'characters_stat_history'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(results)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return data

    async def weapon_stat(self, results: int = 1,
                          **kwargs: Any) -> List[CensusData]:
        """Return weapon-specific statistics for this character."""
        collection: Final[str] = 'characters_weapon_stat'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(results)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return data

    async def weapon_stat_by_faction(self, results: int = 1,
                                     **kwargs: Any) -> List[CensusData]:
        """Return per-faction weapon statistics for this character."""
        collection: Final[str] = 'characters_weapon_stat_by_faction'
        query = Query(collection, service_id=self._client.service_id, **kwargs)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(results)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return data

    def title(self) -> InstanceProxy[Title]:
        """Return the current title of the character, if any.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        title_id = self.data.title_id or -1
        query = Query(Title.collection, service_id=self._client.service_id)
        query.add_term(field=Title.id_field, value=title_id)
        return InstanceProxy(Title, query, client=self._client)

    def world(self) -> InstanceProxy[World]:
        """Return the world of the character.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        collection: Final[str] = 'characters_world'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        join = query.create_join(World.collection)
        join.set_fields('world_id')
        return InstanceProxy(World, query, client=self._client)



"""Low-level Census API wrapper and URL generator.

This module is responsible for generating the URLs used to interface
with the Census API. It is game agnostic and should work for any title
supporting the Daybreak Game Company Census API.
It mostly serves as an internal utility to generate the queries used by
Auraxium's primary, object-oriented interface. It is however completely
stand-alone and can be used on its own.

This module does not perform any queries; it only dynamically generates
the URLs required. Use an HTTP library of your choice to handle the
network side.
When selecting an HTTP library, keep in mind that some API collections,
like ``character`` and ``outfit_member``, can have access times in
excess of several seconds.
For bots and most other use-cases, it is therefore advisable to use an
asynchronous library like `aiohttp`_ to prevent your program from
locking up while waiting for the server to respond.

To generate a census query URL using this module, instantiate a
:class:`Query`, tweak its settings as desired and finally call the
:meth:`Query.url()` method. This will return a :class:`yarl.URL`
instance that can be used as-is or cast to :class:`str`.

Queries come in two flavours, :class:`Query` for top-level queries, and
and :class:`JoinedQuery` for inner, joined queries (aka. joins). Both
inherit from :class:`QueryBase`, which defines shared utility present
in both sub classes.

You can also use one query as a template when creating another, which
can be helpful when building large, deeply nested queries. See the
:meth:`QueryBase.copy()` method for details.

Example:
    This snippet will look up a character by name and return their
    online status.

    .. code-block:: python3

        from auraxium import census

        query = census.Query('character')
        query.add_term(field='name.first_lower', value='auroram')
        join = query.create_join('characters_online_status')
        join.inject_at = 'online_status'
        print(query.url())

    :meth:`Query.url()` returns a :class:`yarl.URL` instance, which you
    would generally pass to your HTTP library of choice, rather than
    printing it.

.. _aiohttp:
    https://docs.aiohttp.org/en/stable/

"""

from .query import JoinedQuery, Query, QueryBase
from .support import SearchModifier, SearchTerm

__all__ = [
    'JoinedQuery',
    'Query',
    'QueryBase',
    'SearchModifier',
    'SearchTerm'
]

__version__ = '0.1.0'



"""Supporting classes and helper methods for the census module."""

import dataclasses
import enum
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union

__all__ = [
    'CensusValue',
    'JoinedQueryData',
    'QueryBaseData',
    'QueryData',
    'SearchModifier',
    'SearchTerm'
]

CensusValue = Union[float, int, str]
# This list connects the string literals to the enum values. The index of the
# list element must match the corresponding enum value.
conversion_list: List[str] = ['', '<', '[', '>', ']', '^', '*', '!']
QueryBaseDataT = TypeVar('QueryBaseDataT', bound='QueryBaseData')


class SearchModifier(enum.Enum):
    """Enumerates the search modifiers available for Term objects.

    Note that a given search modifier may not be valid for all fields.

    The following is a list of all search modifier literals and their
    corresponding enum value:::

        EQUAL_TO: '',               LESS_THAN: '<',
        LESS_THAN_OR_EQUAL: '[',    GREATER_THAN: '>',
        GREATER_THAN_OR_EQUAL: ']', STARTS_WITH: '^',
        CONTAINS: '*',              NOT_EQUAL: '!'

    """

    EQUAL_TO = 0
    LESS_THAN = 1
    LESS_THAN_OR_EQUAL = 2
    GREATER_THAN = 3
    GREATER_THAN_OR_EQUAL = 4
    STARTS_WITH = 5
    CONTAINS = 6
    NOT_EQUAL = 7

    @classmethod
    def from_value(cls, value: CensusValue) -> 'SearchModifier':
        """Infer the search modifier from a given value.

        If the input is a string, its first character will be compared
        against the corresponding API literals. If a match is found,
        the corresponding SearchModifier enum value is returned.

        If the input is not a string or its first character does not
        match any API literal, this will return ``EQUAL_TO``.

        Arguments:
            value: A value to infer the search modifier from.

        Returns:
            The search modifier for the value provided.

        """
        # Return EQUAL_TO for non-string values
        if not isinstance(value, str):
            return cls(cls.EQUAL_TO)
        # For strings, return the corresponding enum value
        try:
            return cls(conversion_list.index(value[0]))
        except ValueError:
            return cls(cls.EQUAL_TO)

    @staticmethod
    def serialise(enum_value: Union[int, 'SearchModifier']) -> str:
        """Return the string literal for the given enum value.

        This is mostly used during URL generation.

        Arguments:
            enum_value: The enum value or index to serialise.

        Raises:
            ValueError: Raised if the provided integer exceeds the
                value range of the enum.

        Returns:
            The string representation of the search modifier. This will
            be an empty string for ``SearchModifier.EQUAL_TO``.

        """
        # Convert the enum value to an integer
        if isinstance(enum_value, SearchModifier):
            enum_value = enum_value.value
        assert isinstance(enum_value, int)
        # Return the appropriate string literal
        try:
            return conversion_list[enum_value]
        except IndexError as err:
            raise ValueError(f'Invalid enum value {enum_value}') from err


class SearchTerm:
    """Represents a single query term."""

    def __init__(self, field: str, value: CensusValue,
                 modifier: SearchModifier = SearchModifier.EQUAL_TO) -> None:
        """Initialise a new search term.

        Search terms are used to filter the results before returning.
        This is particularly important for lists returned by joined
        queries, as they do not have access to limiting mechanisms like
        Query, easily resulting in excessively long return lists.

        Use the :meth:`SearchTerm.infer()` factory if you prefer
        defining search modifiers via their string literals as used by
        the API, rather than manually specifying the enum value.

        Arguments:
            field: The field to compare.
            value: The value to compare the field against.
            modifier(optional): The search modifier to use. Modifiers
                can be used to get non-exact or partial matches.
                Defaults to ``SearchModifier.EQUAL_TO``.

        """
        self.field = field
        self.value = value
        self.modifier = modifier

    def as_tuple(self) -> Tuple[str, str]:
        """Return a key/value pair representing the search term.

        This is a helper function that calls
        :meth:`SearchTerm.serialise()` and then splits the returned
        string at the equal sign.

        Returns:
            A key/value pair representing the search term.

        """
        key, value = self.serialise().split('=', 1)
        return key, value

    @classmethod
    def infer(cls, field: str, value: CensusValue) -> 'SearchTerm':
        """Infer a term from the given field and value.

        This is a more natural way of defining search terms for users
        familiar with the API literals. See the docstring of the
        :class:`SearchModifier` enumerator for a list of search
        modifiers.

        Note that this requires the value to be a str instance; this
        can obscure the actual field value (this is generally not of
        concern as this information will be lost during URL generation
        regardless).

        Arguments:
            field: The field to compare.
            value: The value to compare the field against. If a string,
                it will be checked for a search modifier literal.

        Returns:
            A new SearchTerm with a pre-defined search modifier.

        """
        term = cls(field=field, value=value)
        term.modifier = SearchModifier.from_value(value)
        # If a search modifier other than EQUAL_TO was returned, the
        # SearchModifier.from_value() function has found a literal in the
        # string.
        # This must now be cut from the value string.
        if term.modifier != SearchModifier.EQUAL_TO:
            assert isinstance(value, str)
            term.value = value[1:]
        return term

    def serialise(self) -> str:
        """Return the literal representation of the search term.

        This is the string that will added to the URL's query string as
        part of the URL generator.

        Returns:
            The string representation of the search term.

        """
        return (f'{self.field}={SearchModifier.serialise(self.modifier)}'
                f'{self.value}')


@dataclasses.dataclass()
class QueryBaseData:
    """A dataclass used to store generic query information."""

    collection: Optional[str]
    hide: List[str] = dataclasses.field(default_factory=list)
    joins: List['JoinedQueryData'] = dataclasses.field(default_factory=list)
    show: List[str] = dataclasses.field(default_factory=list)
    terms: List[SearchTerm] = dataclasses.field(default_factory=list)

    @classmethod
    def from_base(cls: Type[QueryBaseDataT],
                  data: 'QueryBaseData') -> QueryBaseDataT:
        """Helper used to copy the base query data components."""
        return cls(**data.__dict__)


@dataclasses.dataclass()
class QueryData(QueryBaseData):
    """A dataclass used to store global flags and settings for queries."""

    case: bool = True
    distinct: Optional[str] = None
    exact_match_first: bool = False
    has: List[str] = dataclasses.field(default_factory=list)
    include_null: bool = False
    lang: Optional[str] = None
    limit: int = 1
    limit_per_db: int = 1
    namespace: str = 'ps2:v2'
    resolve: List[str] = dataclasses.field(default_factory=list)
    retry: bool = True
    service_id: str = 's:example'
    sort: List[Union[str, Tuple[str, bool]]] = dataclasses.field(
        default_factory=list)
    start: int = 0
    timing: bool = False
    tree: Optional[Dict[str, Any]] = None


@dataclasses.dataclass()
class JoinedQueryData(QueryBaseData):
    """Data class for joined queries in the API."""

    inject_at: Optional[str] = None
    is_list: bool = False
    is_outer: bool = True
    field_on: Optional[str] = None
    field_to: Optional[str] = None



"""Experience and rank class definitions."""

import logging
from typing import List, TYPE_CHECKING, Union

import pydantic

from ..base import Cached
from ..client import Client
from ..errors import PayloadError
from ..models import ExperienceData, ExperienceRankData
from ..types import CensusData

if TYPE_CHECKING:
    from ..ps2 import Faction

log = logging.getLogger('auraxium.ps2')


class Experience(Cached, cache_size=100, cache_ttu=3600.0):
    """A type of experience tick."""

    collection = 'experience'
    data: ExperienceData
    dataclass = ExperienceData
    id_field = 'experience_id'


class ExperienceRank:
    """A type of experience tick."""

    collection = 'experience_rank'
    data: ExperienceRankData
    dataclass = ExperienceRankData

    def __init__(self, data: CensusData, client: Client) -> None:
        """Initialise the object.

        This populates the object using the provided payload.
        """
        rank = int(data['rank'])
        log.debug('Instantiating <%s:%d> using payload: %s',
                  self.__class__.__name__, rank, data)
        self._client = client
        try:
            self.data = ExperienceRankData(**data)
        except pydantic.ValidationError as err:
            raise PayloadError(
                f'Unable to populate {self.__class__.__name__} due to a '
                f'missing key: {err.args[0]}', data) from err

    def image(self, faction: Union[int, 'Faction']) -> str:
        """Return the default image for this type."""
        from ..ps2 import Faction  # pylint: disable=import-outside-toplevel
        if isinstance(faction, Faction):
            faction = faction.id
        internal_tag: List[str] = ['null', 'vs', 'nc', 'tr', 'nso']
        image_id = getattr(self.data, internal_tag[faction])
        url = 'https://census.daybreakgames.com/files/ps2/images/static/'
        return url + f'{image_id}.png'

    def __repr__(self) -> str:
        """Return the unique string representation of this object.

        This will take the form of <Class:rank:type>, e.g.
        <ExperienceRank:50:ASP>.

        Returns:
            A string representing the object.

        """
        mode = 'ASP' if self.data.nc.image_id == 88685 else 'Default'
        return f'<{self.__class__.__name__}:{self.data.rank}:{mode}>'



"""Directive class definitions."""

from typing import Final

from ..base import ImageMixin, Named
from ..census import Query
from ..models import (DirectiveData, DirectiveTierData,
                      DirectiveTreeCategoryData, DirectiveTreeData)
from ..proxy import InstanceProxy, SequenceProxy

from .objective import Objective


class DirectiveTreeCategory(Named, ImageMixin, cache_size=10, cache_ttu=300.0):
    """A directive category.

    Directive tree category are the topmost directive categorisation,
    e.g. "Infantry".
    """

    collection = 'directive_tree_category'
    data: DirectiveTreeCategoryData
    dataclass = DirectiveTreeCategoryData
    id_field = 'directive_tree_category_id'

    def trees(self) -> SequenceProxy['DirectiveTree']:
        """Return the trees in the category.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        query = Query(
            DirectiveTree.collection, service_id=self._client.service_id)
        query.add_term(field='directive_tree_id', value=self.id).limit(50)
        return SequenceProxy(DirectiveTree, query, client=self._client)


class DirectiveTree(Named, ImageMixin, cache_size=30, cache_ttu=60.0):
    """A tree of directives.

    Directive trees are a chain of related directive, e.g.
    "Heavy Assault".
    """

    collection = 'directive_tree'
    data: DirectiveTreeData
    dataclass = DirectiveTreeData
    id_field = 'directive_tree_id'

    def category(self) -> InstanceProxy[DirectiveTreeCategory]:
        """Return the category of the directive tree.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(DirectiveTreeCategory.collection,
                      service_id=self._client.service_id)
        query.add_term(field=DirectiveTreeCategory.id_field,
                       value=self.data.directive_tree_category_id)
        return InstanceProxy(DirectiveTreeCategory, query, client=self._client)

    def directives(self) -> SequenceProxy['Directive']:
        """Return the list of directives in this category.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        query = Query(Directive.collection, service_id=self._client.service_id)
        query.add_term(field='directive_tree_id', value=self.id).limit(400)
        return SequenceProxy(Directive, query, client=self._client)

    def tiers(self) -> SequenceProxy['DirectiveTier']:
        """Return the list of directive tiers in this category.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        query = Query(
            DirectiveTier.collection, service_id=self._client.service_id)
        query.add_term(field='directive_tree_id', value=self.id).limit(4)
        return SequenceProxy(DirectiveTier, query, client=self._client)


class DirectiveTier(Named, ImageMixin, cache_size=30, cache_ttu=60.0):
    """A directive tier.

    Container for related directives, e.g. "Combat Medic: Adept".
    """

    collection = 'directive_tier'
    data: DirectiveTierData
    dataclass = DirectiveTierData
    id_field = 'directive_tier_id'

    def directives(self) -> SequenceProxy['Directive']:
        """Return the list of directives in this tier.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        query = Query(Directive.collection, service_id=self._client.service_id)
        query.add_term(field='directive_tier_id', value=self.id).limit(100)
        return SequenceProxy(Directive, query, client=self._client)

    def tree(self) -> InstanceProxy[DirectiveTree]:
        """Return the tree of the directive.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(
            DirectiveTree.collection, service_id=self._client.service_id)
        query.add_term(
            field=DirectiveTree.id_field, value=self.data.directive_tree_id)
        return InstanceProxy(DirectiveTree, query, client=self._client)


class Directive(Named, ImageMixin, cache_size=30, cache_ttu=60.0):
    """A directive a character may complete."""

    collection = 'directive'
    data: DirectiveData
    dataclass = DirectiveData
    id_field = 'directive_id'

    def objectives(self) -> SequenceProxy[Objective]:
        """Return the objectives for this directive.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        # NOTE: This table is being treated as a single set mapping to multiple
        # objectives via their common group. This is a guess. I was not able to
        # find any directives with multiple objectives associated.
        collection: Final[str] = 'objective_set_to_objective'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(
            field='objective_set_id', value=self.data.objective_set_id)
        join = query.create_join(Objective.collection)
        join.set_fields('objective_group_id')
        join.set_list(True)
        return SequenceProxy(Objective, query, client=self._client)

    def tier(self) -> InstanceProxy[DirectiveTier]:
        """Return the tier of the directive.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(
            DirectiveTier.collection, service_id=self._client.service_id)
        query.add_term(
            field=DirectiveTier.id_field, value=self.data.directive_tier_id)
        return InstanceProxy(DirectiveTier, query, client=self._client)

    def tree(self) -> InstanceProxy[DirectiveTree]:
        """Return the tree of the directive.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(
            DirectiveTree.collection, service_id=self._client.service_id)
        query.add_term(
            field=DirectiveTree.id_field, value=self.data.directive_tree_id)
        return InstanceProxy(DirectiveTree, query, client=self._client)



"""Data classes for :mod:`auraxium.ps2.vehicle`."""

from typing import Optional

from ..base import ImageData, Ps2Data
from ..types import LocaleData

__all__ = [
    'VehicleAttachmentData',
    'VehicleData'
]


class VehicleAttachmentData(Ps2Data):
    """Data class for :class:`auraxium.ps2.vehicle.VehicleAttachment`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        item_id: The item that is being attached.
        vehicle_id: The vehicle the item may be attached to.
        faction_id: The faction for which this attachment is available.
        description: A description of the attachment.
        slot_id: The slot the attachment goes into.

    """

    item_id: int
    vehicle_id: int
    faction_id: int
    description: str
    slot_id: int


class VehicleData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.vehicle.Vehicle`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        vehicle_id: The unique ID of this vehicle.
        name: The localised name of the vehicle.
        description: The localised description of the vehicle.
        type_id: The type of vehicle.
        type_name: The name of the type of vehicle.
        cost: The cost of the vehicle.
        cost_resource_id: The ID of the resource the cost is in.

    """

    vehicle_id: int
    name: LocaleData
    description: Optional[LocaleData] = None
    type_id: int
    type_name: str
    cost: Optional[int] = None
    cost_resource_id: Optional[int] = None



# type: ignore
"""Package deployment script."""

import setuptools

with open('README.md') as readme:
    long_description = readme.read()

setuptools.setup(name='auraxium',
                 version='0.1.0a3',
                 author='Leonhard S.',
                 author_email='leonhard-sei@outlook.com',
                 description='A python wrapper for the PlanetSide 2 Census API.',
                 long_description=long_description,
                 long_description_content_type='text/markdown',
                 keywords='auraxium python daybreak census planetside ps2',
                 url='https://github.com/leonhard-s/auraxium',
                 packages=setuptools.find_packages(),
                 package_data={'auraxium': ['py.typed']},
                 classifiers=['Development Status :: 3 - Alpha',
                              'Programming Language :: Python :: 3.8',
                              'License :: OSI Approved :: MIT License',
                              'Operating System :: OS Independent'],
                 install_requires=['aiohttp',
                                   'backoff',
                                   'pydantic',
                                   'yarl',
                                   'websockets'],
                 license='MIT',
                 include_package_data=True,
                 zip_safe=False)



"""Data classes for :mod:`auraxium.ps2.metagame`."""

from typing import Optional

from ..base import Ps2Data
from ..types import LocaleData

__all__ = [
    'MetagameEventData'
]


class MetagameEventData(Ps2Data):
    """Data class for :class:`auraxium.ps2.metagame.MetagameEvent`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        metagame_event_id: The unique ID of this event.
        name: The localised name of the event.
        description: The localised description of the event.
        type: The type of event. Legacy alerts like "Dome Domination"
            used to share the same type despite being different events
            due to taking place on different continents.
        experience_bonus: The experience bonus applied to players in
            percent.

    """

    metagame_event_id: int
    name: LocaleData
    description: LocaleData
    type: int
    experience_bonus: Optional[int] = None



"""Resistance mapping class definitions."""

from ..base import Cached
from ..census import Query
from ..models import ResistInfoData, ResistTypeData
from ..proxy import InstanceProxy


class ResistType(Cached, cache_size=100, cache_ttu=60.0):
    """A type of resistance a profile may hold.

    This is used to implement weapon types like "Melee", "Small Arms"
    or "Heavy Machine Gun".
    """

    collection = 'resist_type'
    data: ResistTypeData
    dataclass = ResistTypeData
    id_field = 'resist_type_id'


class ResistInfo(Cached, cache_size=100, cache_ttu=60.0):
    """Specifies the resistance values for a given profile and type."""

    collection = 'resist_info'
    data: ResistInfoData
    dataclass = ResistInfoData
    id_field = 'resist_info_id'

    def type(self) -> InstanceProxy[ResistType]:
        """Return the resist type for this entry.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(
            ResistType.collection, service_id=self._client.service_id)
        query.add_term(
            field=ResistType.id_field, value=self.data.resist_type_id)
        return InstanceProxy(ResistType, query, self._client)



"""Data classes for :mod:`auraxium.ps2.projectile`."""

from typing import Optional

from ..base import Ps2Data

__all__ = [
    'ProjectileData'
]


class ProjectileData(Ps2Data):
    """Data class for :class:`auraxium.ps2.projectile.Projectile`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        projectile_id: The unique ID of this projectile.
        projectile_flight_type_id: The ID of the associated
            :class:`ProjectileFlightType`.
        speed: The projectile speed in meters per second.
        speed_max: The maximum speed of the projectile; used with
            :attr:`acceleration`.
        acceleration: The acceleration of the projectile in meters
            per second per second.
        turn_rate: The turn rate of the projectile in degrees per
            second.
        lifespan: The time after which the projectile will be
            destroyed if it has not hit a target. Effectively limits
            the maximum distance travelled.
        drag: The drag applied to the projectile.
        gravity: The affect on gravity on the projectile.
        lockon_acceleration: (Not yet documented)
        lockon_lifespan: The life span of the projectile after it has
            locked onto a target.
        arm_distance: The minimum distance travelled for the projectile
            to be armed. Used to increase the Archer's damage on range.
        tether_distance: (Not yet documented)
        detonate_distance: The travelled distance after which the
            projectile will detonate.
        detonate_on_contact: Whether the projectile will explode when
            hitting a target. False for grenades and other bouncing
            projectiles.
        sticky: Whether the projectile will stick to objects.
        sticks_to_players: Whether the projectile will stick to
            players.
        lockon_lose_angle: The angle at which the projectile will lose
            its lock if the target's turn speed exceeds its own.
        lockon_seek_in_flight: Whether the projectile will continue to
            follow its target after launch.

    """

    projectile_id: int
    projectile_flight_type_id: int
    speed: int
    speed_max: Optional[int] = None
    acceleration: Optional[int] = None
    turn_rate: Optional[int] = None
    lifespan: Optional[float] = None
    drag: Optional[float] = None
    gravity: Optional[float] = None
    lockon_acceleration: Optional[float] = None
    lockon_lifespan: Optional[float] = None
    arm_distance: Optional[float] = None
    tether_distance: Optional[float] = None
    detonate_distance: Optional[float] = None
    detonate_on_contact: Optional[bool] = None
    sticky: Optional[bool] = None
    sticks_to_players: Optional[bool] = None
    lockon_lose_angle: Optional[int] = None
    lockon_seek_in_flight: Optional[bool] = None



"""Facility and map class definitions."""

from typing import Final, Set

from ..base import Cached, Named
from ..census import Query
from ..models import FacilityTypeData, MapHexData, MapRegionData, RegionData
from ..proxy import InstanceProxy, SequenceProxy

from .zone import Zone


class FacilityType(Cached, cache_size=10, cache_ttu=3600.0):
    """A type of base/facility found in the game."""

    collection = 'facility_type'
    data: FacilityTypeData
    dataclass = FacilityTypeData
    id_field = 'facility_type_id'


class MapHex(Cached, cache_size=100, cache_ttu=60.0):
    """An individual territory hex in the map."""

    collection = 'map_hex'
    data: MapHexData
    dataclass = MapHexData
    id_field = 'map_region_id'

    def map_region(self) -> InstanceProxy['MapRegion']:
        """Return the map region associated with this map hex."""
        query = Query(MapRegion.collection, service_id=self._client.service_id)
        query.add_term(field=MapRegion.id_field, value=self.data.map_region_id)
        return InstanceProxy(MapRegion, query, client=self._client)


class MapRegion(Cached, cache_size=100, cache_ttu=60.0):
    """A facility on the continent map."""

    collection = 'map_region'
    data: MapRegionData
    dataclass = MapRegionData
    id_field = 'map_region_id'

    async def get_connected(self) -> Set['MapRegion']:
        """Return the facilities connected to this region."""
        # NOTE: This operation cannot be done in a single query as there is no
        # "or" operator.
        collection: Final[str] = 'facility_link'
        connected: Set['MapRegion'] = set()  # type: ignore
        # Set up the base query
        query = Query(collection, service_id=self._client.service_id)
        query.limit(10)
        join = query.create_join(self.collection)
        join.set_fields('facility_id_a', 'facility_id')
        join = query.create_join(self.collection)
        join.set_fields('facility_id_b', 'facility_id')
        # Modified query A
        query.add_term(field='facility_id_a', value=self.data.facility_id)
        proxy: SequenceProxy[MapRegion] = SequenceProxy(
            MapRegion, query, client=self._client)
        connected.update(await proxy.flatten())
        # Modified query B
        query.data.terms = []
        query.add_term(field='facility_id_b', value=self.data.facility_id)
        proxy: SequenceProxy[MapRegion] = SequenceProxy(
            MapRegion, query, client=self._client)
        connected.update(await proxy.flatten())
        return connected

    def zone(self) -> InstanceProxy[Zone]:
        """Return the zone/continent of the region.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Zone.collection, service_id=self._client.service_id)
        query.add_term(field=Zone.id_field, value=self.data.zone_id)
        return InstanceProxy(Zone, query, client=self._client)


class Region(Named, cache_size=100, cache_ttu=60.0):
    """A map region or facility."""

    collection = 'region'
    data: RegionData
    dataclass = RegionData
    id_field = 'region_id'

    def map_region(self) -> InstanceProxy[MapRegion]:
        """Return the map region associated with this region.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(MapRegion.collection, service_id=self._client.service_id)
        query.add_term(field=MapRegion.id_field, value=self.id)
        return InstanceProxy(MapRegion, query, client=self._client)

    def zone(self) -> InstanceProxy[Zone]:
        """Return the zone/continent of the region.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Zone.collection, service_id=self._client.service_id)
        query.add_term(field=Zone.id_field, value=self.data.zone_id)
        return InstanceProxy(Zone, query, client=self._client)



"""Data classes for :mod:`auraxium.ps2.item`."""

from typing import Optional

from ..base import ImageData, Ps2Data
from ..types import LocaleData

__all__ = [
    'ItemCategoryData',
    'ItemData',
    'ItemTypeData'
]


class ItemCategoryData(Ps2Data):
    """Data class for :class:`auraxium.ps2.item.ItemCategory`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        item_category_id: The unique ID of this item category.
        name: The localised name of the category.

    """

    item_category_id: int
    name: LocaleData


class ItemData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.item.Item`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        item_id: The unique ID of this item.
        item_type_id: The ID of the item type for this item.
        item_category_id: The ID of the item category for this item.
        activatable_ability_id: (Not yet documented)
        passive_ability_id: (Not yet documented)
        is_vehicle_weapon: Whether this item is a vehicle weapon.
        name: Localised name of the item.
        description: Localised description of the item.
        faction_id: The faction that has access to this item.
        max_stack_size: The stack size for stackable items such as
            grenades.
        skill_set_id: The skill set associated with this item. This is
            used for upgradable items like the Medical Applicator or
            Repair Tool.
        is_default_attachment: Default attachments are generally not
            visible to the user and are used whenever nothing is
            selected. Examples include the default iron sights, or the
            regular ammo type for weapon supporting non-standard
            ammo types.

    """

    item_id: int
    item_type_id: Optional[int] = None
    item_category_id: Optional[int] = None
    activatable_ability_id: Optional[int] = None
    passive_ability_id: Optional[int] = None
    is_vehicle_weapon: bool
    name: LocaleData
    description: Optional[LocaleData] = None
    faction_id: Optional[int] = None
    max_stack_size: int
    skill_set_id: Optional[int] = None
    is_default_attachment: bool


class ItemTypeData(Ps2Data):
    """Data class for :class:`auraxium.ps2.item.ItemType`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        item_type_id: The unique ID of this item type.
        name: The identifying name of this item type.
        code: The internal code used to describe this item type.

    """

    item_type_id: int
    name: str
    code: str



"""Data classes for :mod:`auraxium.ps2.ability`."""

from typing import Optional

from ..base import Ps2Data

__all__ = [
    'AbilityData',
    'AbilityTypeData',
    'ResourceTypeData'
]


class AbilityData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.Ability`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        ability_id: The unique ID of this ability.
        ability_type_id: The associated ability type for this ability.
        expire_msec: The duration of the ability.
        first_use_delay_msec: The initial cooldown of the ability.
        next_use_delay_msec: The reuse cooldown of the ability.
        resource_type_id: The resource type used by the ability.
        resource_first_cost: The initial cast cost of the ability.
        resource_cost_per_msec: The resource cost per second for
            toggled abilities.
        distance_max: (Not yet documented)
        radius_max: (Not yet documented)
        flag_toggle: Whether the ability is toggled.
        param*: Type-specific parameters for this ability. Refer to the
            corresponding :class:`AbilityType` for details.
        string*: Type-specific string values for this ability. Refer to
            the corresponding :class:`AbilityType` for details.

    """

    ability_id: int
    ability_type_id: int
    expire_msec: Optional[int] = None
    first_use_delay_msec: Optional[int] = None
    next_use_delay_msec: Optional[int] = None
    reuse_delay_msec: Optional[int] = None
    resource_type_id: Optional[int] = None
    resource_first_cost: Optional[int] = None
    resource_cost_per_msec: Optional[int] = None
    distance_max: Optional[float] = None
    radius_max: Optional[float] = None
    flag_toggle: Optional[bool] = None
    param1: Optional[str] = None
    param2: Optional[str] = None
    param3: Optional[str] = None
    param4: Optional[str] = None
    param5: Optional[str] = None
    param6: Optional[str] = None
    param7: Optional[str] = None
    param8: Optional[str] = None
    param9: Optional[str] = None
    param10: Optional[str] = None
    param11: Optional[str] = None
    param12: Optional[str] = None
    param13: Optional[str] = None
    param14: Optional[str] = None
    string1: Optional[str] = None
    string2: Optional[str] = None
    string3: Optional[str] = None
    string4: Optional[str] = None


class AbilityTypeData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.AbilityType`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        ability_type_id: The unique ID for this ability type.
        description: A description of what this ability type is used
            for.
        param*: Descriptions of what the corresponding parameter is
            used for in abilities of this type.
        string*: Descriptions of what the corresponding string value is
            used for in abilities of this type.

    """

    ability_type_id: int
    description: Optional[str] = None
    param1: Optional[str] = None
    param2: Optional[str] = None
    param3: Optional[str] = None
    param4: Optional[str] = None
    param5: Optional[str] = None
    param6: Optional[str] = None
    param7: Optional[str] = None
    param8: Optional[str] = None
    param9: Optional[str] = None
    param10: Optional[str] = None
    param11: Optional[str] = None
    param12: Optional[str] = None
    param13: Optional[str] = None
    param14: Optional[str] = None
    string1: Optional[str] = None
    string2: Optional[str] = None
    string3: Optional[str] = None
    string4: Optional[str] = None


class ResourceTypeData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.ResourceType`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        resource_type_id: The unique ID of this resource type.
        description: A description of what this resource type is used
            for.

    """

    resource_type_id: int
    description: str



"""Data classes for :mod:`auraxium.ps2.world`."""

from typing import Optional

from ..base import Ps2Data
from ..types import LocaleData

__all__ = [
    'WorldData'
]


class WorldData(Ps2Data):
    """Data class for :class:`auraxium.ps2.world.World`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        world_id: The unique ID of the world.
        state: The current state (i.e. online status) of the world.
        name: The localised name of the world.
        description: A description of the world's server region.

    """

    world_id: int
    state: str
    name: LocaleData
    description: Optional[LocaleData] = None



"""Armor mapping class definitions."""

import enum

from ..base import Cached
from ..models import ArmourInfoData


class ArmourFacing(enum.IntEnum):
    """Enumerator for armour facing directions.

    This is used to list different armour values for a vehicle based on
    the angle of attack.
    """

    FRONT = 0
    RIGHT = 1
    TOP = 2
    REAR = 3
    LEFT = 4
    BOTTOM = 5
    ALL = 6


class ArmourInfo(Cached, cache_size=100, cache_ttu=60.0):
    """An armour stat for a given entity.

    Note that any given entity may have multiple :class:`ArmourInfo`
    instances associated with it, one for each :class:`ArmourFacing`
    value.
    """

    collection = 'armor_info'
    data: ArmourInfoData
    dataclass = ArmourInfoData
    id_field = 'armor_info_id'

    @property
    def facing(self) -> ArmourFacing:
        """Return the facing direction for this stat entry."""
        return ArmourFacing(self.data.armor_info_id)



"""Defines the main client for Auraxium.

This includes both the core methods used to interface with the REST
API, as well as the websocket client used to access the real-time event
streaming service (ESS).

"""

import asyncio
import copy
import logging
from typing import Any, List, Literal, Optional, Type, TYPE_CHECKING, TypeVar
from types import TracebackType

import aiohttp

from .census import Query
from .request import run_query
from .types import CensusData

if TYPE_CHECKING:
    # This is only imported during static type checking to resolve the forward
    # references. This avoids a circular import at runtime.
    from .base import Named, Ps2Object

__all__ = [
    'Client'
]

ClientT = TypeVar('ClientT', bound='Client')
NamedT = TypeVar('NamedT', bound='Named')
Ps2ObjectT = TypeVar('Ps2ObjectT', bound='Ps2Object')
log = logging.getLogger('auraxium.client')


class Client:
    """The main client used to interface with the PlanetSide 2 API.

    This class handles access to the REST API at
    ``https://census.daybreakgames.com/``.

    To interface with the REST API, use the methods :meth:`Client.get`,
    :meth:`Client.find()`, or one of the ``Client.get_by_*()`` helpers.

    Attributes:
        loop: The :mod:`asyncio` event loop used by the client.
        service_id: The service ID identifying your app to the API. You
            can use the default value of ``'s:example'``, but you will
            likely run into rate limits. You can sign up for your own
            service ID at http://census.daybreakgames.com/#devSignup.
        session: The :class:`aiohttp.ClientSession` used for REST API
            requests.

    """

    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None,
                 service_id: str = 's:example', profiling: bool = False
                 ) -> None:
        """Initialise a new Auraxium client.

        If loop is not specified, it will be retrieved is using
        :meth:`asyncio.get_event_loop()`.

        Arguments:
            loop (optional): A pre-existing event loop to use for the
                client. Defaults to ``None``.
            service_id (optional): The unique, private service ID of
                the client. Defaults to ``'s:example'``.
            profiling (optional): Whether to enable query and socket
                profiling.

        """
        self.loop = loop or asyncio.get_event_loop()
        self.profiling = profiling
        self._timing_cache: List[float] = []
        self.service_id = service_id
        self.session = aiohttp.ClientSession()

    async def __aenter__(self: ClientT) -> ClientT:
        """Enter the context manager and return the client."""
        return self

    async def __aexit__(self, exc_type: Optional[Type[BaseException]],
                        exc_value: Optional[BaseException],
                        traceback: Optional[TracebackType]) -> Literal[False]:
        """Exit the context manager.

        This closes the internal HTTP session before exiting, no error
        handling will be performed.

        Arguments:
            exc_type: The type of exception that was raised.
            exc_value: The exception value that was raised.
            traceback: The traceback type of the exception.

        Returns:
            Always False, i.e. no error suppression.

        """
        await self.close()
        return False  # Do not suppress any exceptions

    @property
    def latency(self) -> float:
        """Return the average request latency for the client.

        This averages up to the last 100 query times. Use the logging
        utility to gain more insight into which queries take the most
        time.

        """
        if not self._timing_cache:
            return -1.0
        return sum(self._timing_cache) / len(self._timing_cache)

    async def close(self) -> None:
        """Shut down the client.

        This will end the HTTP session used for requests to the REST
        API.

        Call this to clean up before the client object is destroyed.

        """
        log.info('Shutting down client')
        await self.session.close()

    async def count(self, type_: Type[Ps2ObjectT], **kwargs: Any) -> int:
        """Return the number of items matching the given terms.

        Arguments:
            type_: The object type to search for.
            **kwargs: Any number of filters to apply.

        Returns:
            The number of entries entries.

        """
        return await type_.count(client=self, **kwargs)

    async def find(self, type_: Type[Ps2ObjectT], results: int = 10,
                   offset: int = 0, promote_exact: bool = False,
                   check_case: bool = True, **kwargs: Any) -> List[Ps2ObjectT]:
        """Return a list of entries matching the given terms.

        This returns up to as many entries as indicated by the results
        argument. Note that it may be fewer.

        Arguments:
            type_: The object type to search for.
            results (optional): The maximum number of results. Defaults
                to ``10``.
            offset (optional): The number of entries to skip. Useful
                for paginated views. Defaults to ``0``.
            promote_exact (optional): If enabled, exact matches to
                non-exact searches will always come first in the return
                list. Defaults to ``False``.
            check_case (optional): Whether to check case when comparing
                strings. Note that case-insensitive searches are much
                more expensive. Defaults to ``True``.
            **kwargs: Any number of filters to apply.

        Returns:
            A list of matching entries.

        """
        return await type_.find(results=results, offset=offset,
                                promote_exact=promote_exact,
                                check_case=check_case, client=self, **kwargs)

    async def get(self, type_: Type[Ps2ObjectT], check_case: bool = True,
                  **kwargs: Any) -> Optional[Ps2ObjectT]:
        """Return the first entry matching the given terms.

        Like :meth:`Client.find()`, but will only return one item.

        Arguments:
            type_: The object type to search for.
            check_case (optional): Whether to check case when comparing
                strings. Note that case-insensitive searches are much
                more expensive. Defaults to ``True``.
            **kwargs: Any number of filters to apply.

        Returns:
            The first matching entry, or ``None`` if not found.

        """
        return await type_.get(check_case=check_case, client=self, **kwargs)

    async def get_by_id(self, type_: Type[Ps2ObjectT], id_: int
                        ) -> Optional[Ps2ObjectT]:
        """Retrieve an object by its unique Census ID.

        Like :meth:`Client.get()`, but checks the local cache before
        performing the query.

        Arguments:
            type_: The object type to search for.
            id_: The unique ID of the object.

        Returns:
            The entry with the matching ID, or None if not found.

        """
        return await type_.get_by_id(id_, client=self)

    async def get_by_name(self, type_: Type[NamedT], name: str, *,
                          locale: str = 'en') -> Optional[NamedT]:
        """Retrieve an object by its unique name.

        Depending on the ``type_`` specified, this may retrieve a
        cached object, rather than querying the API.

        Keep in mind that not all :class:`Named` objects have a
        localised name; the ``locale`` argument has no effect in these
        cases.

        This query is always case-insensitive.

        Arguments:
            type_: The object type to search for.
            name: The name to search for.
            locale (optional): The locale of the search key. Defaults
                to ``'en'``.

        Returns:
            The entry with the matching name, or ``None`` if not found.

        """
        return await type_.get_by_name(name, locale=locale, client=self)

    async def request(self, query: Query, verb: str = 'get') -> CensusData:
        """Perform a REST API request.

        This performs the query and performs error checking to ensure
        the query is valid.

        Refer to the :meth:`auraxium.request.raise_for_dict()` method
        for a list of exceptions raised from API errors.

        Arguments:
            query: The query to perform.
            verb (optional): The query verb to utilise.
                Defaults to ``'get'``.

        Returns:
            The API response payload received.

        """
        if self.profiling:
            # Create a copy of the query before modifying it
            query = copy.copy(query)
            query.timing(True)
        data = await run_query(query, verb=verb, session=self.session)
        if self.profiling and verb == 'get':
            timing = data.pop('timing')
            if log.level <= logging.DEBUG:
                url = query.url()
                log.debug('Query times for "%s?%s": %s',
                          '/'.join(url.parts[-2:]), url.query_string,
                          ', '.join([f'{k}: {v}' for k, v in timing.items()]))
            self._timing_cache.append(float(timing['total-ms']))
        return data
